unit chargraf;

{
************************************************************

chargraf.pas - unit for chart.pas

(c) Heiko Purnhagen 1993-2000
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/purnhage/

This program may be copied according to the GNU General Public Licence.
See file COPYING for details.

950805   born
960725   using setwritemode for revline
960726   added revcircle
970729   moved stuff from chart.pas here
970803   up/down for editchar/editline, shiftpressed
980421   moved more stuff from chart.pas here
980614   clkdelay
980623   updatefinecore
980729   drawlines
980811   fixed name display @ length
990824   added revlinethick drawmarkthick
000116   startfinexy

************************************************************
}


interface

uses
  graph,crt,dos,
  mousedrv,serial,grafbmp,chartran;

const
  mouseratio = 4;   { mouse fine speed-ratio }

  fastratio = 20;      { mouse fast stepsize }

  curup = #72;     { up }
  curdown = #80;   { down }
  curleft = #75;   { left }
  curright = #77;  { right }

  keyup = #5;      { ^E }
  keydown = #24;   { ^X }
  keyleft = #19;   { ^S }
  keyright = #4;   { ^D }

  maxmenu = 10;

  tx1 = 2;
  tx2 = 1;
  ty1 = 2;
  ty2 = 1;
  txb = 2;
  twb = 10;
  lowcur = 1;

  wincol = 70;
  winrow = 9;

  esc = #27;
  ctrly = #25;
  backsp = #8;
  cr = #13;
  tab = #9;
  cdel = #83;
  cins = #82;

  cup = #72;
  cdown = #80;
  cleft = #75;
  cright = #77;
  cpgup = #73;
  cpgdn = #81;
  chome = #71;
  cend = #79;

  { ctrl & cursor }
  ctrlleft = #115;
  ctrlright = #116;
  ctrlpgup = #132;
  ctrlpgdn = #118;
  ctrlhome = #119;
  ctrlend = #117;

  { fkt-keys: f1=#59 .. f10=#68 }
  fkt = 58;
  shiftfkt = 83;
  ctrlfkt = 93;
  altfkt = 103;


  circleseg4 = 4;

  mousedel = 20;

  notdisplayed = 32000.0;
  clipfact = 10;
  ms = 1;   { 1/6 mark size }
  landdot = 4;   { distance to dotted line "on land" }

  filecomment = '. chart';

  maxgridlines = 50;
  defscalelimitfact = 10;
  gpsheadposfact = 0.2;


type
  string20 = string[20];
  string80 = string[80];

  menut = record
            x1,y1,x2,y2 : integer;
            vertical : boolean;
            enum : integer;
            eheight : integer;
            ewidth : array [1..maxmenu] of integer;
            key : array [1..maxmenu] of char;
            name : array [1..maxmenu] of string20;
            background : pointer;
          end;

  coordtypet = (cline,clineend,cpoly,cpolyend,ctext,cmark);
  filetypet = (fcha,fovl,flog,fmem,fwpt);

  coordptr = ^coordt;
  blockptr = ^blockt;
  coordt = record
             ll : llt;
             xy : xyt;
             uv : uvt;
             name : blockptr;
             coordtype : coordtypet;
             pred,succ : coordptr;
           end;
  blockt = record
             name : string80;
             scalelimit : real;
             typec : char;
             coord : coordptr;
             filetype : filetypet;
             pred,succ : blockptr;
           end;



var
  xmax,ymax : integer;
  xasp,yasp : word;
  reverse : boolean;
  bgc,fgc : integer;
  mousefine : boolean;
  mouseslow : boolean;
  cursorbig : boolean;
  usemouse : boolean;

  mx,my,mxa,mya : integer;
  mbl,mbr,mbla,mbra : boolean;
  mblold,mbrold : boolean;
  cursorar : array [0..3] of cursort;
  cxi,cyi : integer;
  cx,cy : real;
  mmouseratio : integer;
  mmousefine : boolean;
  mkey : boolean;
  mc1,mc2 : char;

  menum : menut;
  menumaxwidth,menumaxsize : integer;

  th,tw : integer;
  winx1,winy1,winx2,winy2 : integer;
  winxc : integer;
  winyr : array [1..winrow] of integer;
  winbackground : pointer;
  confirmbackground : pointer;

  redraw,recalc : boolean;

  circlesin : array[0..circleseg4*4-1] of real;

  coordtypename : array [coordtypet] of string20;
  filetypename : array [filetypet] of string20;

  coordfirst,coordlast : coordptr;
  blockfirst,blocklast : blockptr;

  displayvector : boolean;
  displaytext : boolean;
  displayovl : boolean;
  displayscalelimit : boolean;
  displaytextlen : integer;
  displaychartsize : real;
  griddist : real;

  currentchartsize : real;
  currentchartcenter : llt;
  defscalelimit : real;

  bmpdisp,bmpdispact,bmpdisplast : boolean;

  ll : llt;
  xy : xyt;
  uv : uvt;

  top1x,top2x,top3x,top4x : integer;
  bot1x,bot2x,bot3x,bot4x,bot5x,bot6x : integer;

  gpsll : llt;
  gpscog,gpssog,gpsdist,gpsbrg : real;
  gpsmsg : string;
  gpsheadtime : real;

  bmpbuf : array [1..2] of bmpbufp;
  bmpxyfact : array [1..2] of real;
  bmpxyrot : array [1..2] of integer;
  bmpxoff,bmpyoff,bmpx0,bmpy0,bmpxs,bmpys : array [1..2] of integer;


procedure revpixel (x,y : integer);
procedure revline (x1,y1,x2,y2 : integer);
procedure revlinethick (x1,y1,x2,y2 : integer);
procedure revcircleinit;
procedure revcircle (x,y,rx,ry : integer);
procedure drawmarkdot (x,y : integer);
procedure drawmarkcircle (x,y : integer);
procedure drawmarkthick (x,y : integer);
procedure newscreen;
procedure box (x1,y1,x2,y2 : integer);
procedure revbox (x1,y1,x2,y2 : integer);
procedure errorbeep;

procedure initfinecolor (cursoron : boolean);
procedure initfine;
procedure xhidecursor;
procedure xshowcursor;
procedure startfinexy (x,y : real);
procedure startfine;
function updatefinecore (usecurkeys : boolean) : boolean;
{ any mouse or key action -> true }
procedure updatefine (usecurkeys : boolean);
procedure endfine;
function clickedl : boolean;
function clickedr : boolean;
function anyclick : boolean;
function getmkey : char;
function clickedlkey : boolean;
function clickedrkey : boolean;
procedure waitclickkey;

procedure newtext (x,y : integer; xcent : boolean; s : string);
procedure boxtext (x1,y1,x2,y2 : integer; xcent : boolean; s : string);
function editchar (x,y : integer; que : string; var ans : char) : integer;
{ <ret>,<key> -> 0, <down> -> 1, <up> -> -1, <esc> -> -9999 }
function editline (x,y : integer; que : string; var ans : string) : integer;
{ <ret> -> 0, <down> -> 1, <up> -> -1, <esc> -> -9999 }
function str2int (s : string; var i : integer) : boolean;
{ OK->true error->false }
function str2real (s : string; var r : real) : boolean;
{ OK->true error->false }

procedure initm (x,y : integer; vert : boolean);
procedure nextm (c : char; s : string20);
procedure quitm (var mm : menut);
procedure drawmenu (var m : menut);
function checkmenu (var m : menut) : integer;
{ ord(mc1) or ord(mc2)+256 if mc1=#0 }
procedure freemenu (var m : menut);
procedure initconfirm (minleft : longint);
function confirm (s : string; yesno : boolean) : boolean;
procedure initwindow (minleft : longint);
procedure openwindow;
procedure closewindow;

function selectfile (filetypename : string; actmask : string;
                     var actfile : string) : boolean;
{ OK->true escape->false }
function showfile (name : string) : boolean;
{ OK->true error->false }
function checkoverwrite (name : string) : boolean;
{ overwrite->true abort->false }

procedure initcoord;
procedure transcoord;
function round2xy (uv : uvt; var x,y : integer) : boolean;
procedure drawgrid;
procedure drawlines;
procedure drawmark (x,y : integer; filetype : filetypet);
procedure drawmarks;
procedure getdrawtext (coordtype : coordtypet; s : string; c : char;
                       var x,y : integer);
procedure drawtexts;

procedure removecoord (coord : coordptr);
procedure removeblock (block : blockptr);
procedure insertcoord (coord,after : coordptr);
{ after = nil -> insert at the begin ... }
procedure insertblock (block,after : blockptr);
procedure cleard (filetype : filetypet);
function loadd (filename : string; filetype : filetypet;
                minmem : longint) : boolean;
function saved (filename : string; filetype : filetypet;
                minmem : longint) : boolean;

function choosegrid (dist : real) : real;
procedure setchartsize;
procedure chartfindab (chartcenter : llt; chartwidth : real);

procedure showtop (left : boolean; s : string);
procedure newactpos;
procedure initactpos;
procedure showmode (s : string);
function brg2str (brg : real) : string;
procedure newgpscur;
procedure newgpsdata (onlymsg : boolean);
procedure initgpsdata;
procedure gpsheadpos (var llh : llt);

procedure initbitmap (minleft : longint);
procedure calcbitmaps2b (bn : integer; xs,ys : real; var xb,yb : real);
procedure calcbitmapb2s (bn : integer; xb,yb : real; var xs,ys : real);
function loadbitmap (fn : string; xcenter,ycenter,xsize,ysize : integer;
                     zoom,thr,rot : integer; bn : integer) : boolean;
{ OK->true error->false }
{ zoom=-9999 -> fit into size}
function showbitmap (bn : integer; bmpcol : integer) : boolean;
function savebitmap (fn : string) : boolean;
function loadbmc (bmcfn : string; var bmpfn : string; var thr,rot : integer;
                  var llar : llart; var uvar : uvart) : boolean;
function savebmc (bmcfn : string; bmpfn : string; thr,rot : integer;
                  llar : llart; uvar : uvart) : boolean;


implementation

{ reverse line with bresenham }

procedure revpixel (x,y : integer);
begin
  putpixel(x,y,getmaxcolor-getpixel(x,y));
end;

{
procedure revline1 (x,y,dbig,dsmall,dir : integer);
  var
    i,j,d : integer;
begin
  if dir > 4 then
    d := (dbig+1) div 2
  else
    d := dbig div 2;
  j := 0;
  for i := 0 to dbig do
  begin
    case dir of
      1 : revpixel (x+i,y+j);
      2 : revpixel (x+j,y+i);
      3 : revpixel (x-j,y+i);
      4 : revpixel (x-i,y+j);
      5 : revpixel (x-i,y-j);
      6 : revpixel (x-j,y-i);
      7 : revpixel (x+j,y-i);
      8 : revpixel (x+i,y-j);
    end;
    d := d + dsmall;
    if d >= dbig then
    begin
      inc(j);
      d := d - dbig;
    end;
  end;
end;
}

{ somehow buggy ... (vert. line len=2)   HP 960725 }
{
procedure revline_old (x1,y1,x2,y2 : integer);
  var
    dx,dy : integer;
begin
  dx := x2-x1;
  dy := y2-y1;
  if dx >= 0 then
  begin
    if dy >= 0 then
    begin
      if dx >= dy then
        revline1 (x1,y1,dx,dy,1)
      else
        revline1 (x1,y1,dy,dx,2);
    end
    else
    begin
      if dx >= -dy then
        revline1 (x1,y1,dx,-dy,8)
      else
        revline1 (x1,y1,-dy,dx,7);
    end;
  end
  else
  begin
    if dy >= 0 then
    begin
      if -dx >= dy then
        revline1 (x1,y1,-dx,dy,4)
      else
        revline1 (x1,y1,dy,-dx,3);
    end
    else
    begin
      if -dx >= -dy then
        revline1 (x1,y1,-dx,-dy,5)
      else
        revline1 (x1,y1,-dy,-dx,6);
    end;
  end;
end;
}

procedure revline (x1,y1,x2,y2 : integer);
  var
    oldcolor : integer;
begin
  setwritemode (xorput);
  oldcolor := getcolor;
  setcolor (getmaxcolor);
  line (x1,y1,x2,y2);
  setcolor (oldcolor);
  setwritemode (normalput);
end;

procedure revlinethick (x1,y1,x2,y2 : integer);
  var
    dx,dy : integer;
  function sign (x : integer) : integer;
  begin
    if x = 0 then
      sign := 0
    else if x < 0 then
      sign := -1
    else
      sign := 1;
  end;
begin
  dx := sign(y2-y1);
  dy := sign(x2-x1);
{
  if ((x2-x1)*dy) div 2 > (y2-y1)*dx then
    dx := 0
  else if (x2-x1)*dy < ((y2-y1)*dx) div 2 then
    dy := 0;
}
  if (x2-x1)*dy > (y2-y1)*dx then
    dx := 0
  else
    dy := 0;
  revline (x1-dx,y1+dy,x2-dx,y2+dy);
  revline (x1,y1,x2,y2);
  revline (x1+dx,y1-dy,x2+dx,y2-dy);
end;

{$ifdef oldstuff}
procedure revlinethick (x1,y1,x2,y2 : integer);
begin
  setlinestyle (solidln,0,thickwidth);
  revline (x1,y1,x2,y2);
  setlinestyle (solidln,0,normwidth);
end;
{$endif}

procedure revcircleinit;
  var i : integer;
begin
  for i := 0 to circleseg4*4-1 do
    circlesin[i] := sin(i/(circleseg4*4)*2*pi);
end;

procedure revcircle (x,y,rx,ry : integer);
  var
    oldcolor : integer;
    i : integer;
begin
  setwritemode (xorput);
  oldcolor := getcolor;
  setcolor (getmaxcolor);
  moveto (x+round(rx*circlesin[circleseg4*4-1]),
          y+round(ry*circlesin[circleseg4-1]));
  for i := 0 to circleseg4*4-1 do
  begin
    linerel (0,0);
    lineto (x+round(rx*circlesin[i]),
            y+round(ry*circlesin[(circleseg4+i) mod (circleseg4*4)]));
  end;
  setcolor (oldcolor);
  setwritemode (normalput);
end;

procedure drawmarkdot (x,y : integer);
begin
  revline (x-1,y-2,x+1,y-2);
  revline (x-2,y-1,x+2,y-1);
  revline (x-2,y,x+2,y);
  revline (x-2,y+1,x+2,y+1);
  revline (x-1,y+2,x+1,y+2);
end;

procedure drawmarkcircle (x,y : integer);
  procedure circlequad (a,b1,b2 : integer);
  begin
    revline (x+b1,y+a,x+b2,y+a);
    revline (x+a,y-b1,x+a,y-b2);
    revline (x-b1,y-a,x-b2,y-a);
    revline (x-a,y+b1,x-a,y+b2);
  end;
begin
  circlequad (6,-4,-3);
  circlequad (8,-2,2);
  circlequad (7,-4,4);
  circlequad (6,3,6);
  circlequad (5,5,6);
end;

procedure drawmarkthick (x,y : integer);
  procedure circlequad (a,b1,b2 : integer);
  begin
    revline (x+b1,y+a,x+b2,y+a);
    revline (x+a,y-b1,x+a,y-b2);
    revline (x-b1,y-a,x-b2,y-a);
    revline (x-a,y+b1,x-a,y+b2);
  end;
begin
{
  circlequad (9,-5,-3);
  circlequad (12,-3,3);
  circlequad (11,-5,6);
  circlequad (10,-5,7);
  circlequad (9,3,8);
  circlequad (8,6,9);
  circlequad (7,7,10);
  circlequad (6,8,11);
}
  circlequad (9,-5,-4);
  circlequad (12,-4,4);
  circlequad (11,-5,6);
  circlequad (10,-5,7);
  circlequad (9,4,8);
  circlequad (8,6,9);
  circlequad (7,7,10);
  circlequad (6,8,11);
end;


{ procedures for mouse cursor }

function shiftpressed : boolean;
  var
    r : registers;
begin
  r.ah := 2;
  intr ($16,r);
  shiftpressed := (r.al and 3) <> 0;
end;

procedure setcursorar;

  procedure setcurline (var cur : cursort; line : integer; s : string);

  { cursor[line] cursor[lin+16]  code  color       }
  { 0            0               0     black       }
  { 1            0               _     transparent }
  { 0            1               1     white       }
  { 1            1               x     invert      }

    var
      i : integer;
      w : word;
  begin
    cur[line] := 0;
    cur[line+16] := 0;
    w := 1;
    for i := 16 downto 1 do
    begin
      case s[i] of
        '0' : if reverse then
                cur[line+16] := cur[line+16]+w;
        '_' : cur[line] := cur[line]+w;
        '1' : if not reverse then
                cur[line+16] := cur[line+16]+w;
        'x' : begin
                cur[line] := cur[line]+w;
                cur[line+16] := cur[line+16]+w;
              end;
      end;
      w := w*2;
    end;
  end;

begin
  if cursorbig then
  begin
    setcurline (cursorar[0], 0,'_____00000______');
    setcurline (cursorar[0], 1,'_____01110______');
    setcurline (cursorar[0], 2,'_____01110______');
    setcurline (cursorar[0], 3,'_____01110______');
    setcurline (cursorar[0], 4,'_____01110______');
    setcurline (cursorar[0], 5,'000000000000000_');
    setcurline (cursorar[0], 6,'011110___011110_');
    setcurline (cursorar[0], 7,'011110___011110_');
    setcurline (cursorar[0], 8,'011110___011110_');
    setcurline (cursorar[0], 9,'000000000000000_');
    setcurline (cursorar[0],10,'_____01110______');
    setcurline (cursorar[0],11,'_____01110______');
    setcurline (cursorar[0],12,'_____01110______');
    setcurline (cursorar[0],13,'_____01110______');
    setcurline (cursorar[0],14,'_____00000______');
    setcurline (cursorar[0],15,'________________');
    setcurline (cursorar[1], 0,'_____000000_____');
    setcurline (cursorar[1], 1,'_____011110_____');
    setcurline (cursorar[1], 2,'_____011110_____');
    setcurline (cursorar[1], 3,'_____011110_____');
    setcurline (cursorar[1], 4,'_____011110_____');
    setcurline (cursorar[1], 5,'0000000000000000');
    setcurline (cursorar[1], 6,'011110____011110');
    setcurline (cursorar[1], 7,'011110____011110');
    setcurline (cursorar[1], 8,'011110____011110');
    setcurline (cursorar[1], 9,'0000000000000000');
    setcurline (cursorar[1],10,'_____011110_____');
    setcurline (cursorar[1],11,'_____011110_____');
    setcurline (cursorar[1],12,'_____011110_____');
    setcurline (cursorar[1],13,'_____011110_____');
    setcurline (cursorar[1],14,'_____000000_____');
    setcurline (cursorar[1],15,'________________');
    setcurline (cursorar[2], 0,'_____00000______');
    setcurline (cursorar[2], 1,'_____01110______');
    setcurline (cursorar[2], 2,'_____01110______');
    setcurline (cursorar[2], 3,'_____01110______');
    setcurline (cursorar[2], 4,'_____01110______');
    setcurline (cursorar[2], 5,'000000000000000_');
    setcurline (cursorar[2], 6,'011110___011110_');
    setcurline (cursorar[2], 7,'011110___011110_');
    setcurline (cursorar[2], 8,'011110___011110_');
    setcurline (cursorar[2], 9,'011110___011110_');
    setcurline (cursorar[2],10,'000000000000000_');
    setcurline (cursorar[2],11,'_____01110______');
    setcurline (cursorar[2],12,'_____01110______');
    setcurline (cursorar[2],13,'_____01110______');
    setcurline (cursorar[2],14,'_____01110______');
    setcurline (cursorar[2],15,'_____00000______');
    setcurline (cursorar[3], 0,'_____000000_____');
    setcurline (cursorar[3], 1,'_____011110_____');
    setcurline (cursorar[3], 2,'_____011110_____');
    setcurline (cursorar[3], 3,'_____011110_____');
    setcurline (cursorar[3], 4,'_____011110_____');
    setcurline (cursorar[3], 5,'0000000000000000');
    setcurline (cursorar[3], 6,'011110____011110');
    setcurline (cursorar[3], 7,'011110____011110');
    setcurline (cursorar[3], 8,'011110____011110');
    setcurline (cursorar[3], 9,'011110____011110');
    setcurline (cursorar[3],10,'0000000000000000');
    setcurline (cursorar[3],11,'_____011110_____');
    setcurline (cursorar[3],12,'_____011110_____');
    setcurline (cursorar[3],13,'_____011110_____');
    setcurline (cursorar[3],14,'_____011110_____');
    setcurline (cursorar[3],15,'_____000000_____');
  end
  else
  begin
    setcurline (cursorar[0], 0,'______000_______');
    setcurline (cursorar[0], 1,'______010_______');
    setcurline (cursorar[0], 2,'______010_______');
    setcurline (cursorar[0], 3,'______010_______');
    setcurline (cursorar[0], 4,'______010_______');
    setcurline (cursorar[0], 5,'______000_______');
    setcurline (cursorar[0], 6,'000000___000000_');
    setcurline (cursorar[0], 7,'011110___011110_');
    setcurline (cursorar[0], 8,'000000___000000_');
    setcurline (cursorar[0], 9,'______000_______');
    setcurline (cursorar[0],10,'______010_______');
    setcurline (cursorar[0],11,'______010_______');
    setcurline (cursorar[0],12,'______010_______');
    setcurline (cursorar[0],13,'______010_______');
    setcurline (cursorar[0],14,'______000_______');
    setcurline (cursorar[0],15,'________________');
    setcurline (cursorar[1], 0,'______0000______');
    setcurline (cursorar[1], 1,'______0110______');
    setcurline (cursorar[1], 2,'______0110______');
    setcurline (cursorar[1], 3,'______0110______');
    setcurline (cursorar[1], 4,'______0110______');
    setcurline (cursorar[1], 5,'______0000______');
    setcurline (cursorar[1], 6,'000000____000000');
    setcurline (cursorar[1], 7,'011110____011110');
    setcurline (cursorar[1], 8,'000000____000000');
    setcurline (cursorar[1], 9,'______0000______');
    setcurline (cursorar[1],10,'______0110______');
    setcurline (cursorar[1],11,'______0110______');
    setcurline (cursorar[1],12,'______0110______');
    setcurline (cursorar[1],13,'______0110______');
    setcurline (cursorar[1],14,'______0000______');
    setcurline (cursorar[1],15,'________________');
    setcurline (cursorar[2], 0,'______000_______');
    setcurline (cursorar[2], 1,'______010_______');
    setcurline (cursorar[2], 2,'______010_______');
    setcurline (cursorar[2], 3,'______010_______');
    setcurline (cursorar[2], 4,'______010_______');
    setcurline (cursorar[2], 5,'______000_______');
    setcurline (cursorar[2], 6,'000000___000000_');
    setcurline (cursorar[2], 7,'011110___011110_');
    setcurline (cursorar[2], 8,'011110___011110_');
    setcurline (cursorar[2], 9,'000000___000000_');
    setcurline (cursorar[2],10,'______000_______');
    setcurline (cursorar[2],11,'______010_______');
    setcurline (cursorar[2],12,'______010_______');
    setcurline (cursorar[2],13,'______010_______');
    setcurline (cursorar[2],14,'______010_______');
    setcurline (cursorar[2],15,'______000_______');
    setcurline (cursorar[3], 0,'______0000______');
    setcurline (cursorar[3], 1,'______0110______');
    setcurline (cursorar[3], 2,'______0110______');
    setcurline (cursorar[3], 3,'______0110______');
    setcurline (cursorar[3], 4,'______0110______');
    setcurline (cursorar[3], 5,'______0000______');
    setcurline (cursorar[3], 6,'000000____000000');
    setcurline (cursorar[3], 7,'011110____011110');
    setcurline (cursorar[3], 8,'011110____011110');
    setcurline (cursorar[3], 9,'000000____000000');
    setcurline (cursorar[3],10,'______0000______');
    setcurline (cursorar[3],11,'______0110______');
    setcurline (cursorar[3],12,'______0110______');
    setcurline (cursorar[3],13,'______0110______');
    setcurline (cursorar[3],14,'______0110______');
    setcurline (cursorar[3],15,'______0000______');
  end;
end;

procedure revcursor (mx,my : integer);
begin
  revline (mx-7,my-1,mx-4,my-1);
  revline (mx-7,my,mx-4,my);
  revline (mx-7,my+1,mx-4,my+1);
  revline (mx+4,my-1,mx+7,my-1);
  revline (mx+4,my,mx+7,my);
  revline (mx+4,my+1,mx+7,my+1);
  revline (mx-1,my-7,mx-1,my-4);
  revline (mx,my-7,mx,my-4);
  revline (mx+1,my-7,mx+1,my-4);
  revline (mx-1,my+4,mx-1,my+7);
  revline (mx,my+4,mx,my+7);
  revline (mx+1,my+4,mx+1,my+7);
end;

procedure limitborder (var mx,my : integer);
begin
  if mx < 0 then
    mx := 0;
  if mx > xmax then
    mx := xmax;
  if my < 0 then
    my := 0;
  if my > ymax then
    my := ymax;
end;

procedure initfinecolor (cursoron : boolean);
begin
  if not usemouse then
    exit;
  if cursoron then
    hidecursor;
  setcursorar;
  cursor := cursorar[0];
  defgcursor (7,7);
  showcursor;
end;

procedure initfine;
begin
  cx := (xmax+1)/2;
  cy := ty1+ty2+textheight(' ')+5+10;
  mx := round(cx);
  my := round(cy);
  if usemouse then
  begin
    setminmaxx (0,xmax);
    setminmaxy (0,ymax);
    setpos (mx,my);
    initfinecolor (false);
    getposbut (mx,my,mbl,mbr);
    mbla := mbl;
    mbra := mbr;
    mblold := mbla;
    mbrold := mbra;
    mmouseratio := 1;
    cxi := 0;
    cyi := 0;
  end
  else
    revcursor (mx,my);
  mxa := mx;
  mya := my;
  cxi := mx;
  cyi := my;
  mkey := false;
  mmousefine := false;
end;

procedure xhidecursor;
begin
  if usemouse then
    hidecursor
  else
    revcursor (mx,my);
end;

procedure xshowcursor;
begin
  if usemouse then
    showcursor
  else
    revcursor (mx,my);
end;

procedure updatefinecalc;
  var
    cxii,cyii : integer;
begin
  if usemouse and mmousefine then
  begin
    cxi := cxi + (mx - mxa);
    cyi := cyi + (my - mya);
    cxii := cxi div mmouseratio;
    cyii := cyi div mmouseratio;
    cx := cxii / 2;
    cy := cyii / 2;
    mx := cxii div 2;
    my := cyii div 2;
    cursor := cursorar[(cxii mod 2) + 2*(cyii mod 2)];
    hidecursor;
    defgcursor (7,7);
    setpos (mx,my);
    showcursor;
  end
  else
  begin
    cx := mx;
    cy := my;
  end;
end;

procedure startfinexy (x,y : real);
  var
    cxii,cyii : integer;
begin
  mmousefine := mousefine;
  if (not mmousefine) or (not usemouse) then
  begin
    mmouseratio := 1;
    if (x>=0) and (y>=0) and (x<=xmax) and (y<=ymax) then
    begin
      mx := round (x);
      my := round (y);
      limitborder (mx,my);
      if usemouse then
        setpos (mx,my)
      else
      begin
        revcursor (mxa,mya);
        revcursor (mx,my);
      end;
      mxa := mx;
      mya := my;
    end;
    exit;
  end;
  if mouseslow then
    mmouseratio := mouseratio
  else
    mmouseratio := 1;
  if (x>=0) and (y>=0) and
     (x<=xmax*2*mmouseratio) and (y<=ymax*2*mmouseratio) then
  begin
    cxi := trunc ((x*2+0.5)*mmouseratio);
    cyi := trunc ((y*2+0.5)*mmouseratio);
    cxii := cxi div mmouseratio;
    cyii := cyi div mmouseratio;
    cx := cxii / 2;
    cy := cyii / 2;
    mx := cxii div 2;
    my := cyii div 2;
    limitborder (mx,my);
    setpos (mx,my);
    mxa := mx;
    mya := my;
  end;
  if ((cxi div mmouseratio) div 2 <> mx) or
     ((cyi div mmouseratio) div 2 <> my) then
  begin
    cxi := mx*2*mmouseratio;
    cyi := my*2*mmouseratio;
    cx := mx;
    cy := my;
  end;
  updatefinecalc;
end;

procedure startfine;
begin
  startfinexy (-1,-1);
end;

function updatefinecore (usecurkeys : boolean) : boolean;
{ any mouse or key action -> true }
  var
    mstep : integer;
begin
  updatefinecore := false;
  if usemouse then
  begin
    getposbut (mx,my,mbl,mbr);
    if (mxa = mx) and (mya = my) and (mbla = mbl) and (mbra = mbr) and
       not keypressed then
      exit;
  end
  else
    if not keypressed then
      exit;
  updatefinecore := true;
  if keypressed and not mkey then
  begin
    mkey := true;
    mc1 := readkey;
    if mc1 = #0 then
    begin
      mc2 := readkey;
      if usecurkeys and (mc2 in [curup,curdown,curleft,curright]) then
      begin
        mstep := mmouseratio;
        if shiftpressed then
          mstep := mstep*fastratio;
        case mc2 of
          curup    : my := my - mstep;
          curdown  : my := my + mstep;
          curleft  : mx := mx - mstep;
          curright : mx := mx + mstep;
        end;
        limitborder (mx,my);
        if usemouse then
          setpos (mx,my)
        else
        begin
          revcursor (mxa,mya);
          revcursor (mx,my);
        end;
        mkey := false;
      end;
    end;
  end;
  updatefinecalc;
  mxa := mx;
  mya := my;
  if not usemouse then
    exit;
  mblold := mbla;
  mbrold := mbra;
  mbla := mbl;
  mbra := mbr;
end;

procedure updatefine (usecurkeys : boolean);
begin
  repeat
  until updatefinecore (usecurkeys);
end;

procedure endfine;
begin
  if (not mmousefine) or (not usemouse) then
    exit;
  mmousefine := false;
  cursor := cursorar[0];
  hidecursor;
  defgcursor (7,7);
  setpos (mx,my);
  showcursor;
end;

function clickedl : boolean;
begin
  clickedl := usemouse and mbl and not mblold;
end;

function clickedr : boolean;
begin
  clickedr := usemouse and mbr and not mbrold;
end;

function anyclick : boolean;
begin
  anyclick := clickedl or clickedr or mkey;
end;

function getmkey : char;
begin
  getmkey := mc1;
  mkey := false;
end;

function clickedlkey : boolean;
begin
  if mkey then
    clickedlkey := (getmkey = ' ')
  else
    clickedlkey := clickedl;
end;

function clickedrkey : boolean;
begin
  if mkey then
    clickedrkey := (getmkey <> ' ')
  else
    clickedrkey := clickedr;
end;

procedure waitclickkey;
begin
  repeat
    updatefine (true);
  until anyclick;
  mkey := false;
end;


{ procedures for basic graphic & text  i/o }

procedure newscreen;
begin
  xhidecursor;
  setfillstyle (1,bgc);
  bar (0,0,xmax,ymax);
  xshowcursor;
end;

procedure box (x1,y1,x2,y2 : integer);
begin
  moveto (x1,y1);
  lineto (x1,y2);
  lineto (x2,y2);
  lineto (x2,y1);
  lineto (x1,y1);
end;

procedure revbox (x1,y1,x2,y2 : integer);
  var
    t : integer;
begin
  if (x1 = x2) or (y1 = y2) then
    revline (x1,y1,x2,y2)
  else
  begin
    if x2 < x1 then
    begin
      t := x1;
      x1 := x2;
      x2 := t;
    end;
    if y2 < y1 then
    begin
      t := y1;
      y1 := y2;
      y2 := t;
    end;
    revline (x1,y1,x1,y2-1);
    revline (x1,y2,x2-1,y2);
    revline (x2,y2,x2,y1+1);
    revline (x2,y1,x1+1,y1);
  end;
end;

procedure errorbeep;
begin
  sound (1000);
  clkdelay (100);
  nosound;
end;

procedure newtext (x,y : integer; xcent : boolean; s : string);
begin
  xhidecursor;
  if xcent then
    x := x - (textwidth(s)+tx1+tx2) div 2;
  bar (x,y,x+textwidth(s)+tx1+tx2,y+th+ty1+ty2);
  moveto (x+tx1,y+ty1);
  outtext (s);
  xshowcursor;
end;

procedure boxtext (x1,y1,x2,y2 : integer; xcent : boolean; s : string);
begin
  xhidecursor;
  bar (x1,y1,x2,y2);
  box (x1+1,y1+1,x2-1,y2-1);
  if xcent then
    newtext ((x1+x2+1) div 2,y1+3,true,s)
  else
    newtext (x1+3,y1+3,false,s);
  xshowcursor;
end;

function editchar (x,y : integer; que : string; var ans : char) : integer;
{ <ret>,<key> -> 0, <down> -> 1, <up> -> -1, <esc> -> -9999 }
  var
    c : char;
    x2 : integer;
    t : integer;
begin
  x2 := x+textwidth(que)+tx1;
  xhidecursor;
  bar (x,y,x2+tw+tx2,y+th+ty1+ty2);
  moveto (x+tx1,y+ty1);
  outtext (que);
  moveto (x2,y+ty1);
  outtext (ans);
  moveto (x2,y+ty1+lowcur);
  outtext ('_');
  xshowcursor;
  repeat
    t := 9999;
    repeat
      updatefine (false);
    until clickedr or mkey;
    if clickedr then
      t := -9999
    else { mkey }
    begin
      c := getmkey;
      if c = cr then
        t := 0;
      if c = esc then
        t := -9999;
      if c >= ' ' then
      begin
        ans := c;
        t := 0;
      end;
      if c = #0 then
        if mc2 = cdown then
          t := 1
        else
          if mc2 = cup then
            t := -1;
    end;
  until t <> 9999;
  editchar := t;
  xhidecursor;
  bar (x2,y,x2+tw+tx2,y+th+ty1+ty2);
  moveto (x2,y+ty1);
  outtext (ans);
  xshowcursor;
end;

function editline (x,y : integer; que : string; var ans : string) : integer;
{ edit old ans, <ret> to enter new, <esc> to keep old }
{ <ret> -> 0, <down> -> 1, <up> -> -1, <esc> -> -9999 }
  var
    act : integer;
    c : char;
    t : integer;
    old : string;
    x2 : integer;
    w,answ,oldw : integer;
    firstkey : boolean;
begin
  x2 := x+textwidth(que)+tx1;
  xhidecursor;
  bar (x,y,x2,y+th+ty1+ty2);
  moveto (x+tx1,y+ty1);
  outtext (que);
  xshowcursor;
  old := ans;
  act := 1;
  oldw := 0;
  firstkey := true;
  repeat
    t := 9999;
    answ := textwidth(ans)+tw;
    w := answ;
    if oldw > w then
      w := oldw;
    oldw := answ;
    xhidecursor;
    bar (x2,y,x2+w+tx2,y+th+ty1+ty2);
    moveto (x2,y+ty1);
    outtext (ans);
    moveto (x2+(act-1)*tw,y+ty1+lowcur);
    outtext ('_');
    xshowcursor;
    repeat
      updatefine (false);
    until anyclick;
    if clickedl then
    begin
      if (mx >= x2) and (mx <= x2+textwidth(ans)+tw) and
         (my >= y+ty1) and (my <= y+th+ty1) then
      begin
        act := trunc((mx-x2)/tw) + 1;
        if act > length(ans)+1 then
          act := length(ans)+1;
      end;
    end
    else if clickedr then
      t := -9999
    else   { mkey }
    begin
      c := getmkey;
      if c >= ' ' then
      begin
        if firstkey then
        begin
          ans := '';
          act := 1;
        end;
        ans := concat(copy(ans,1,act-1),c,copy(ans,act,length(ans)-act+1));
        inc (act);
      end;
      if c = backsp then
        if act > 1 then
        begin
          ans := concat(copy(ans,1,act-2),copy(ans,act,length(ans)-act+1));
          dec (act);
        end;
      if c = ctrly then
      begin
        ans := '';
        act := 1;
      end;
      if c = cr then
        t := 0;
      if c = esc then
        t := -9999;
      if c = #0 then
      begin
        c := mc2;
        case c of
          cleft :  begin
                     if act > 1 then
                       dec (act);
                   end;
          cright : begin
                     if act < length(ans)+1 then
                       inc (act);
                   end;
          chome : act := 1;
          cend : act := length(ans)+1;
          cdel :   begin
                     ans := concat(copy(ans,1,act-1),
                                   copy(ans,act+1,length(ans)-act));
                   end;
          cup : t := -1;
          cdown : t := 1;
        end;
      end;
    end;
    firstkey := false;
  until t <> 9999;
  if t = -9999 then
    ans := old;
  editline := t;
end;

function str2int (s : string; var i : integer) : boolean;
{ OK->true error->false }
  var
    er : integer;
    tmp : integer;
begin
  str2int := false;
  val (s,tmp,er);
  if er = 0 then
  begin
    str2int := true;
    i := tmp;
  end;
end;

function str2real (s : string; var r : real) : boolean;
{ OK->true error->false }
  var
    er : integer;
    tmp : real;
begin
  str2real := false;
  val (s,tmp,er);
  if er = 0 then
  begin
    str2real := true;
    r := tmp;
  end;
end;


{ menu & window handling procedures }

function lowcase (c : char) : char;
begin
  if (c >= 'A') and (c <= 'Z') then
    lowcase := char(ord(c)+32)
  else
    lowcase := c;
end;

procedure initm (x,y : integer; vert : boolean);
begin
  menum.x1 := x;
  menum.y1 := y;
  menum.x2 := menum.x1;
  menum.vertical := vert;
  menum.enum := 0;
  menumaxwidth := 0;
end;

procedure nextm (c : char; s : string20);
begin
  inc (menum.enum);
  if menum.vertical then
  begin
    if tw*length(s)+tx1+tx2 > menumaxwidth then
      menumaxwidth := tw*length(s)+tx1+tx2;
  end
  else
  begin
    menum.ewidth[menum.enum] := tw*length(s)+tx1+tx2+2*txb;
    menum.x2 := menum.x2 + menum.ewidth[menum.enum];
  end;
  menum.key[menum.enum] := c;
  menum.name[menum.enum] := s;
end;

procedure quitm (var mm : menut);
  var
    i,msize : integer;
begin
  menum.eheight := th+ty1+ty2;
  if menum.vertical then
  begin
    for i := 1 to menum.enum do
      menum.ewidth[i] := menumaxwidth;
    menum.x2 := menum.x1 + menumaxwidth + 5;
    menum.y2 := menum.y1 + menum.eheight*menum.enum + 5;
  end
  else
  begin
    menum.x2 := menum.x2 + 5;
    menum.y2 := menum.y1 + menum.eheight + 5;
  end;
  msize := imagesize(menum.x1,menum.y1,menum.x2,menum.y2);
  if msize > menumaxsize then
    menumaxsize := msize;
  menum.background := nil;
  mm := menum;
end;

procedure drawmenu (var m : menut);
  var
    i,x,y : integer;
begin
  xhidecursor;
  if m.background <> nil then
    getimage (m.x1,m.y1,m.x2,m.y2,m.background^);
  bar (m.x1,m.y1,m.x2,m.y2);
  box (m.x1+1,m.y1+1,m.x2-1,m.y2-1);
  x := m.x1 + 3;
  y := m.y1 + 3;
  for i := 1 to m.enum do
  begin
    if m.vertical then
    begin
      moveto (x+tx1,y+ty1);
      y := y + m.eheight;
    end
    else
    begin
      moveto (x+txb+tx1,y+ty1);
      x := x + m.ewidth[i];
    end;
    outtext (m.name[i]);
  end;
  xshowcursor;
end;

function checkmenu (var m : menut) : integer;
{ ord(mc1) or ord(mc2)+256 if mc1=#0 }
  var
    c : char;
    i,x,y : integer;
begin
  checkmenu := ord(esc);
  if mkey then
  begin
    c := getmkey;
    if c = #0 then
      checkmenu := 256+ord(mc2);
    if c = tab then
      checkmenu := ord(tab)
    else if c = ' ' then
      checkmenu := ord(c)
    else
      for i := 1 to m.enum do
        if lowcase(c) = m.key[i] then
          checkmenu := ord(m.key[i]);
    exit;
  end;
  if clickedr then
    exit;
  checkmenu := ord(' ');   { clickedl }
  x := m.x1 + 3;
  y := m.y1 + 3;
  for i := 1 to m.enum do
  begin
    if (mx >= x) and (mx < x+m.ewidth[i]) and
       (my >= y) and (my < y+m.eheight) then
    begin
      checkmenu := ord(m.key[i]);
      exit;
    end;
    if m.vertical then
      y := y + m.eheight
    else
      x := x + m.ewidth[i];
  end;
end;

procedure freemenu (var m : menut);
begin
  xhidecursor;
  if m.background <> nil then
    putimage (m.x1,m.y1,m.background^,normalput)
  else
  begin
    bar (m.x1,m.y1,m.x2,m.y2);
    redraw := true;
  end;
  xshowcursor;
end;

procedure initconfirm (minleft : longint);
  var
    x1,y1,x2,y2 : integer;
    xs,ys : integer;
    memneed : longint;
begin
  xs := wincol*tw;
  xs := xs+tx1+tx2 + 2*twb;
  ys := 3*(th+ty1+ty2) + 2*twb;
  x1 := (xmax+1) div 2 - (xs+1) div 2;
  y1 := (ymax+1) div 2 - (ys+1) div 2;
  x2 := x1 + xs - 1;
  y2 := y1 + ys - 1;
  memneed := imagesize(x1,y1,x2,y2);
  if memneed+minleft < maxavail then
    getmem (confirmbackground,memneed)
  else
    confirmbackground := nil;
end;

function confirm (s : string; yesno : boolean) : boolean;
  var
    x1,y1,x2,y2 : integer;
    xs,ys : integer;
    xy,xn,yy : integer;
begin
  s := copy(s,1,wincol);
  xs := length(s)*tw;
  if xs < 9*tw then
    xs := 9*tw;
  xs := xs+tx1+tx2 + 2*twb;
  ys := 3*(th+ty1+ty2) + 2*twb;
  x1 := (xmax+1) div 2 - (xs+1) div 2;
  y1 := (ymax+1) div 2 - (ys+1) div 2;
  x2 := x1 + xs - 1;
  y2 := y1 + ys - 1;
  if yesno then
  begin
    xy := xmax div 2 - 3*tw - (3*tw) div 2;
    xn := xmax div 2 + 3*tw - (2*tw) div 2;
  end
  else
    xy := xmax div 2 - (2*tw) div 2;
  yy := y1 + twb + 2*(th+ty1+ty2) + ty1;
  xhidecursor;
  if confirmbackground <> nil then
    getimage (x1,y1,x2,y2,confirmbackground^);
  bar (x1,y1,x2,y2);
  box (x1+1,y1+1,x2-1,y2-1);
  moveto (xmax div 2 - textwidth(s) div 2,y1 + twb + ty1);
  outtext (s);
  if yesno then
  begin
    moveto (xy,yy);
    outtext ('Yes');
    moveto (xn,yy);
    outtext ('No');
  end
  else
  begin
    moveto (xy,yy);
    outtext ('OK');
  end;
  xshowcursor;
  repeat
    updatefine (true);
  until anyclick;
  if yesno then
  begin
    if mkey then
      confirm := (lowcase(getmkey) = 'y')
    else
      if clickedl then
        confirm := ((mx >= xy-tx1) and (mx < xy+3*tw+tx2) and
                    (my >= yy-ty1) and (my < yy+th+ty2))
      else
        confirm := false;   { clickedr }
  end
  else
  begin
    mkey := false;
    confirm := true;
  end;
  xhidecursor;
  if confirmbackground <> nil then
    putimage (x1,y1,confirmbackground^,normalput)
  else
  begin
    bar (x1,y1,x2,y2);
    redraw := true;
  end;
  xshowcursor;
end;

procedure initwindow (minleft : longint);
  var
    w,h,i : integer;
    memneed : longint;
begin
  w := tw*wincol+tx1+tx2+2*twb;
  h := (th+ty1+ty2)*winrow+2*twb;
  winx1 := xmax div 2 - w div 2;
  winy1 := ymax div 2 - h div 2;
  winx2 := winx1 + w - 1;
  winy2 := winy1 + h - 1;
  memneed := imagesize(winx1,winy1,winx2,winy2);
  if memneed+minleft < maxavail then
    getmem (winbackground,memneed)
  else
    winbackground := nil;
  winxc := (winx1 + winx2) div 2;
  for i := 1 to winrow do
    winyr[i] := winy1 + twb + (i-1)*(th+ty1+ty2);
end;

procedure openwindow;
begin
  xhidecursor;
  if winbackground <> nil then
    getimage (winx1,winy1,winx2,winy2,winbackground^);
  bar (winx1,winy1,winx2,winy2);
  box (winx1+1,winy1+1,winx2-1,winy2-1);
  xshowcursor;
end;

procedure closewindow;
begin
  xhidecursor;
  if winbackground <> nil then
    putimage (winx1,winy1,winbackground^,normalput)
  else
  begin
    bar (winx1,winy1,winx2,winy2);
    redraw := true;
  end;
  xshowcursor;
end;

function selectfile (filetypename : string; actmask : string;
                     var actfile : string) : boolean;
{ OK->true escape->false }
  const
    maxdir = 100;
  var
    srec : searchrec;
    dir : array[0..maxdir] of string[14];
    numdir : integer;
    s : string;
    c : char;
    actpath : string;
    b : boolean;
    act,i,j,i0 : integer;
    oldfile : string;
begin
  oldfile := actfile;
  i0 := 0;
  if oldfile = '' then
    i0 := 1;
  actfile := '';
  selectfile := false;
  repeat
    repeat
      i := editline (winx1+twb,winyr[(winrow+1) div 2],
                     filetypename+' filename: ',actmask);
      if i = -9999 then
        exit;
    until i = 0;
    if (pos('*',actmask) > 0) or (pos('?',actmask) > 0) then
    begin
      i := length(actmask);
      while (i > 0) and (actmask[i] <> '\') and (actmask[i] <> ':') do
        dec (i);
      actpath := copy(actmask,1,i);
      numdir := 0;
      findfirst (actmask,0,srec);
      while (doserror = 0) and (numdir < maxdir) do
      begin
        inc (numdir);
        dir[numdir] := srec.name;
        findnext (srec);
      end;
      if numdir = 0 then
        b := confirm ('no files found',false)
      else
      begin
        if numdir >= maxdir then
        begin
          str (maxdir,s);
          b := confirm ('more than '+s+' files found',false);
        end;
        { bubblesort }
        for i := numdir-1 downto 1 do
          for j := 1 to i do
            if dir[j] > dir[j+1] then
            begin
              dir[0] := dir[j];
              dir[j] := dir[j+1];
              dir[j+1] := dir[0];
            end;
        act := i0;
        repeat
          if act = 0 then
            s := oldfile
          else
            s := actpath+dir[act];
          newtext (winx1+twb,winyr[(winrow+1) div 2],false,
                   'select '+filetypename+' file: '+s+
                   '              ');
          repeat
            clkdelay (mousedel);
            updatefine (false);
          until clickedr or mkey;
          i := 0;
          if clickedr then
            i := 1
          else   { mkey }
          begin
            c := getmkey;
            if c = esc then
              i := 1;
            if c = cr then
              i := 2;
            if c = backsp then
              i := 3;
            if (c = ' ') or (c = tab) then
              i := 4;
            if c = #0 then
            begin
              c := mc2;
              if (c = cleft) or (c = cup) then
                i := 3;
              if (c = cright) or (c = cdown) then
                i := 4;
            end;
          end;
          if i = 3 then
          begin
            if act > i0 then
              dec (act)
            else
              errorbeep;
          end;
          if i = 4 then
          begin
            if act < numdir then
              inc (act)
            else
              errorbeep;
          end;
        until (i = 1) or (i = 2);
        if i = 2 then
          actfile := s;
        newtext (winx1+twb,winyr[(winrow+1) div 2],false,
                 '       '+'        '+'       '+'              ');
      end;
    end
    else
      actfile := actmask;
  until actfile <> '';
  selectfile := true;
end;


function showfile (name : string) : boolean;
{ OK->true error->false }
  const
    s0 = '                                   '+
         '                                   ';
  var
    topl,toplold,numl,i : integer;
    f : text;
    b : boolean;
    s : string;
    sa : array [1..winrow] of string[wincol];
    c : char;
begin
  showfile := false;
  openwindow;
  assign (f,name);
  {$i-}
  reset (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm('Error: can''t open file '+name,false);
    closewindow;
    exit;
  end;
  topl := 1;
  toplold := 2;
  repeat
    if topl < 1 then
      topl := 1;
    if topl < toplold then
    begin
      close(f);
      reset (f);
      numl := 0;
    end;
    while (not eof(f)) and (numl < topl+winrow-1) do
    begin
      {$i-}
      readln (f,s);
      {$i+}
      if ioresult <> 0 then
      begin
        b := confirm('Error: can''t read file '+name,false);
        closewindow;
        exit;
      end;
      inc(numl);
      if length(s) > wincol then
        sa[1+(numl mod winrow)] := copy(s,1,wincol)
      else
        sa[1+(numl mod winrow)] := s+copy(s0,1,wincol-length(s));
    end;
    if topl > numl-winrow+1 then
      topl := numl-winrow+1;
    if topl < 1 then
      topl := 1;
    toplold := topl;
    i := topl;
    while (i <= numl) and (i < topl+winrow) do
    begin
      newtext (winx1+twb,winyr[i-topl+1],false,sa[1+(i mod winrow)]);
      inc(i);
    end;
    b := false;
    repeat
      updatefine (false);
    until clickedr or mkey;
    if clickedr then
      b := true
    else { mkey }
    begin
      c := getmkey;
      if c = esc then
        b := true;
      if c = cr then
        topl := topl+wincol;
      if c = backsp then
        topl := topl-1;
      if (c = ' ') or (c = tab) then
        topl := topl+1;
      if c = #0 then
      begin
        c := mc2;
        if c = cup then
          topl := topl-1;
        if c = cdown then
          topl := topl+1;
        if c = cpgup then
          topl := topl-winrow;
        if c = cpgdn then
          topl := topl+winrow;
        if c = chome then
          topl := 1;
        if c = cend then
          topl := 9999;
      end;
    end;
  until b;
  showfile := true;
  close (f);
  closewindow;
end;

function checkoverwrite (name : string) : boolean;
{ overwrite->true abort->false }
  var
    f : file;
begin
  checkoverwrite := true;
  assign (f,name);
  {$i-}
  reset (f);
  {$i+}
  if ioresult = 0 then
  begin
    close (f);
    checkoverwrite := confirm (name+' exists! overwrite?',true);
  end;
end;


{ draw grid and chart etc. }

procedure initcoord;
begin
  coordfirst := nil;
  coordlast := nil;
  blockfirst := nil;
  blocklast := nil;
  coordtypename[cline]    := 'line';
  coordtypename[clineend] := 'line';
  coordtypename[cpoly]    := 'poly';
  coordtypename[cpolyend] := 'poly';
  coordtypename[ctext]    := 'text';
  coordtypename[cmark]    := 'mark';
  filetypename[fcha] := 'chart';
  filetypename[fovl] := 'overlay';
  filetypename[flog] := 'log';
  filetypename[fmem] := 'memo';
  filetypename[fwpt] := 'waypoint';
end;

procedure transcoord;
  var
    coord : coordptr;
begin
  coord := coordfirst;
  while coord <> nil do
  begin
    if ((coord^.name^.filetype = fovl) and not displayovl) or
       ((coord^.name^.scalelimit < displaychartsize) and
        (coord^.name^.scalelimit <> 0) and displayscalelimit) then
    begin
      coord^.uv.u := notdisplayed;
      coord^.uv.v := notdisplayed;
    end
    else
    begin
      ll2xy (coord^.ll,coord^.xy);
      xy2uv (coord^.xy,coord^.uv);
    end;
    coord := coord^.succ;
  end;
end;

function round2xy (uv : uvt; var x,y : integer) : boolean;
begin
  if (uv.u < (1+clipfact)*xmax) and (uv.u > -clipfact*xmax) and
     (uv.v < (1+clipfact)*ymax) and (uv.v > -clipfact*ymax) then
  begin
    x := round(uv.u);
    y := round(uv.v);
    round2xy := true;
  end
  else
  begin
    x := 0;
    y := 0;
    round2xy := false;
  end;
end;

procedure drawgrid;
  var
    minlat,maxlat,minlon,maxlon : real;
    uv1,uv2 : uvt;
    xy1,xy2 : xyt;
    ll1,ll2 : llt;
    raddist : real;
    x1,y1,x2,y2 : integer;

  procedure setminmax;
  begin
    uv2xy (uv1,xy1);
    xy2ll (xy1,ll1);
    if ll1.lat < minlat then
      minlat := ll1.lat;
    if ll1.lat > maxlat then
      maxlat := ll1.lat;
    if ll1.lon < minlon then
      minlon := ll1.lon;
    if ll1.lon > maxlon then
      maxlon := ll1.lon;
  end;

begin
  if griddist = 0 then
    exit;
  uv1.u := 0;
  uv1.v := 0;
  uv2xy (uv1,xy1);
  xy2ll (xy1,ll1);
  minlat := ll1.lat;
  maxlat := ll1.lat;
  minlon := ll1.lon;
  maxlon := ll1.lon;
  uv1.u := xmax;
  setminmax;
  uv1.v := ymax;
  setminmax;
  uv1.u := 0;
  setminmax;
  xhidecursor;
  raddist := griddist/60/180*pi;
  ll1.lat := minlat;
  ll2.lat := maxlat;
  ll1.lon := trunc(minlon/raddist)*raddist;
  while ll1.lon < maxlon do
  begin
    ll2.lon := ll1.lon;
    ll2xy (ll1,xy1);
    xy2uv (xy1,uv1);
    ll2xy (ll2,xy2);
    xy2uv (xy2,uv2);
    if round2xy(uv1,x1,y1) and round2xy(uv2,x2,y2) then
      line (x1,y1,x2,y2);
    ll1.lon := ll1.lon + raddist;
  end;
  ll1.lon := minlon;
  ll2.lon := maxlon;
  ll1.lat := trunc(minlat/raddist)*raddist;
  while ll1.lat < maxlat do
  begin
    ll2.lat := ll1.lat;
    ll2xy (ll1,xy1);
    xy2uv (xy1,uv1);
    ll2xy (ll2,xy2);
    xy2uv (xy2,uv2);
    if round2xy(uv1,x1,y1) and round2xy(uv2,x2,y2) then
      line (x1,y1,x2,y2);
    ll1.lat := ll1.lat + raddist;
  end;
  xshowcursor;
end;

procedure drawlines;
  var
    coord : coordptr;
    x,y,xa,ya,xp,yp,dx,dy : integer;
    df : integer;
    drawing,ispoly,firstpoly,closepoly : boolean;
    c : char;
begin
  xhidecursor;
  drawing := false;
  ispoly := false;
  coord := coordfirst;
  while coord <> nil do
  begin
    if displayovl or (coord^.name^.filetype <> fovl) then
    begin
      if coord^.coordtype in [cline,clineend,cpoly,cpolyend] then
      begin
        c := coord^.name^.typec;
        if (c < 'a') or (c > 'f') then
          c := 'a';
        case c of
          'a','f' : setlinestyle (solidln,0,normwidth);
          'b' : setlinestyle (dottedln,0,normwidth);
          'c' : setlinestyle (centerln,0,normwidth);
          'd' : setlinestyle (dashedln,0,normwidth);
          'e' : setlinestyle (solidln,0,thickwidth);
        end;
        if coord^.coordtype in [cpoly,cpolyend] then
        begin
          firstpoly := not ispoly;
          if firstpoly then
            closepoly := false;
          ispoly := true;
        end;
        if round2xy (coord^.uv,x,y) then
        begin
          if firstpoly then
          begin
            xp := x;
            yp := y;
            closepoly := true;
          end;
          if drawing then
          begin
            line (xa,ya,x,y);
            if c = 'f' then
            begin
              dx := y-ya;
              dy := xa-x;
              df := abs(dx);
              if abs(dy) > df then
                df := abs(dy);
              df := (df+((abs(dx)+abs(dy)) div 2)) div landdot;
              if df > 0 then
              begin
                dx := dx div df;
                dy := dy div df;
                setlinestyle (dottedln,0,normwidth);
                line (xa+dx,ya+dy,x+dx,y+dy);
                setlinestyle (solidln,0,normwidth);
              end;
            end;
          end
          else
            line(x,y,x,y);
          xa := x;
          ya := y;
          drawing := true;
        end
        else
          drawing := false;
      end;
      if (coord^.coordtype = cpolyend) and closepoly and drawing then
      begin
        line (xa,ya,xp,yp);
        if c = 'f' then
        begin
          dx := yp-ya;
          dy := xa-xp;
          df := abs(dx);
          if abs(dy) > df then
            df := abs(dy);
          df := (df+((abs(dx)+abs(dy)) div 2)) div landdot;
          if df > 0 then
          begin
            dx := dx div df;
            dy := dy div df;
            setlinestyle (dottedln,0,normwidth);
            line (xa+dx,ya+dy,xp+dx,yp+dy);
            setlinestyle (solidln,0,normwidth);
          end;
        end;
      end;
      if not (coord^.coordtype in [cline,cpoly]) then
      begin
        drawing := false;
        ispoly := false;
      end;
    end;
    coord := coord^.succ;
  end;
  setlinestyle (solidln,0,normwidth);
  xshowcursor;
end;

procedure drawmark (x,y : integer; filetype : filetypet);
begin
  moveto (x,y);
  case filetype of
    fcha : begin
             moverel (3*ms,0);
             linerel (-6*ms,0);
             moverel (3*ms,3*ms);
             linerel (0,-6*ms);
           end;
    fovl : begin
             moverel (3*ms,0);
             linerel (-6*ms,0);
             moverel (3*ms,3*ms);
             linerel (0,-6*ms);
           end;
    fmem : begin
             moverel (2*ms,2*ms);
             linerel (-4*ms,-4*ms);
             moverel (0,4*ms);
             linerel (4*ms,-4*ms);
           end;
    fwpt : begin
             moverel (3*ms,-1*ms);
             linerel (0,2*ms);
             linerel (-2*ms,2*ms);
             linerel (-2*ms,0);
             linerel (-2*ms,-2*ms);
             linerel (0,-2*ms);
             linerel (2*ms,-2*ms);
             linerel (2*ms,0);
             linerel (2*ms,2*ms);
           end;
  end;
end;

procedure drawmarks;
  var
    coord : coordptr;
    x,y : integer;
begin
  xhidecursor;
  coord := coordfirst;
  while coord <> nil do
  begin
    if displayovl or (coord^.name^.filetype <> fovl) then
      if coord^.coordtype = cmark then
        if round2xy (coord^.uv,x,y) then
          drawmark (x,y,coord^.name^.filetype);
    coord := coord^.succ;
  end;
  xshowcursor;
end;

procedure getdrawtext (coordtype : coordtypet; s : string; c : char;
                       var x,y : integer);
begin
  if (c < 'a') or (c > 'i') then
  begin
    if coordtype = ctext then
      c := 'e'
    else
      c := 'i';
  end;
  if c in ['a','d','g'] then
    x := x-3*ms-2-textwidth(s);
  if c in ['b','e','h'] then
    x := x+1-textwidth(s) div 2;
  if c in ['c','f','i'] then
    x := x+3*ms+4;
  if c in ['a','b','c'] then
    y := y-3*ms-2-textheight(s);
  if c in ['d','e','f'] then
    y := y+1-textheight(s) div 2;
  if c in ['g','h','i'] then
    y := y+3*ms+4;
end;

procedure drawtexts;
  var
    coord : coordptr;
    x,y : integer;
    s : string;
begin
  xhidecursor;
  coord := coordfirst;
  while coord <> nil do
  begin
    if displayovl or (coord^.name^.filetype <> fovl) then
      if coord^.coordtype in [ctext,cmark] then
        if round2xy (coord^.uv,x,y) then
        begin
          s := coord^.name^.name;
          if displaytextlen > 0 then
            s := copy(s,1,displaytextlen);
          getdrawtext (coord^.coordtype,s,coord^.name^.typec,x,y);
          moveto (x,y);
          outtext (s);
        end;
    coord := coord^.succ;
  end;
  xshowcursor;
end;


{ clear, load & save of charts etc. }

procedure removecoord (coord : coordptr);
begin
  if coord^.pred = nil then
    coordfirst := coord^.succ
  else
    coord^.pred^.succ := coord^.succ;
  if coord^.succ = nil then
    coordlast := coord^.pred
  else
    coord^.succ^.pred := coord^.pred;
end;

procedure removeblock (block : blockptr);
begin
  if block^.pred = nil then
    blockfirst := block^.succ
  else
    block^.pred^.succ := block^.succ;
  if block^.succ = nil then
    blocklast := block^.pred
  else
    block^.succ^.pred := block^.pred;
end;

procedure insertcoord (coord,after : coordptr);
{ after = nil -> insert at the begin ... }
begin
  coord^.pred := after;
  if after = nil then
  begin
    coord^.succ := coordfirst;
    if coordfirst = nil then
      coordlast := coord
    else
      coordfirst^.pred := coord;
    coordfirst := coord;
  end
  else
  begin
    coord^.succ := after^.succ;
    if after^.succ = nil then
      coordlast := coord
    else
      after^.succ^.pred := coord;
    after^.succ := coord;
  end;
end;

procedure insertblock (block,after : blockptr);
{ after = nil -> insert at the begin ... }
begin
  block^.pred := after;
  if after = nil then
  begin
    block^.succ := blockfirst;
    if blockfirst = nil then
      blocklast := block
    else
      blockfirst^.pred := block;
    blockfirst := block;
  end
  else
  begin
    block^.succ := after^.succ;
    if after^.succ = nil then
      blocklast := block
    else
      after^.succ^.pred := block;
    after^.succ := block;
  end;
end;

procedure cleard (filetype : filetypet);
  var
    actblock,nextblock : blockptr;
    actcoord,nextcoord : coordptr;
begin
  actblock := blockfirst;
  while actblock <> nil do
  begin
    nextblock := actblock^.succ;
    if actblock^.filetype = filetype then
    begin
      actcoord := actblock^.coord;
      while actcoord^.coordtype in [cline,cpoly] do
      begin
        nextcoord := actcoord^.succ;
        removecoord (actcoord);
        dispose (actcoord);
        actcoord := nextcoord;
      end;
      removecoord (actcoord);
      dispose (actcoord);
      removeblock (actblock);
      dispose (actblock);
    end;
    actblock := nextblock;
  end;
  recalc := true;
end;

function loadd (filename : string; filetype : filetypet;
                minmem : longint) : boolean;
  var
    f : text;
    s,si : string;
    b,found : boolean;
    linenum : integer;

  function readlnok (var s : string; expectline : boolean) : boolean;
    var
      b : boolean;
  begin
    if eof(f) then
    begin
      if expectline then
      begin
        b := confirm ('Error: unexpected end of '+filename,false);
      end;
      readlnok := false;
      exit;
    end;
    {$i-}
    readln (f,s);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t read '+filename,false);
      readlnok := false;
    end
    else
    begin
      inc (linenum);
      readlnok := true;
    end;
  end;

  function loadblock (var s : string; coordtype : coordtypet) : boolean;
    var
      actblock : blockptr;
      actcoord : coordptr;
      scalelimit : real;
      typec : char;
      ll : llt;
      b : boolean;
      er : integer;
  begin
    loadblock := true;
    if copy(s,1,length(coordtypename[coordtype])) <>
       coordtypename[coordtype] then
      exit;
    found := true;
    scalelimit := 0;
    typec := ' ';
    delete (s,1,length(coordtypename[coordtype]));
    while (s <> '') and (s[1] in [' ',tab]) do
      delete (s,1,1);
    if (s <> '') and not (s[1] in ['0'..'9','.']) then
    begin
      typec := s[1];
      delete (s,1,1);
    end;
    while (s <> '') and (s[1] in [' ',tab]) do
      delete (s,1,1);
    while (s <> '') and (s[length(s)] in [' ',tab]) do
      delete (s,length(s),1);
    if s <> '' then
    begin
      val (s,scalelimit,er);
      if er <> 0 then
      begin
        str (linenum,si);
        b := confirm ('Error: can''t read scalelimit in line '+si+' of '+
                      filename,false);
        loadblock := false;
        exit;
      end;
    end;
    if not readlnok (s,true) then
    begin
      loadblock := false;
      exit;
    end;
    if minmem > maxavail then
    begin
      b := confirm ('Error: not enough memory available',false);
      loadblock := false;
      exit;
    end;
    new (actblock);
    actblock^.filetype := filetype;
    actblock^.name := s;
    actblock^.scalelimit := scalelimit;
    actblock^.typec := typec;
    if not readlnok (s,true) then
    begin
      dispose (actblock);
      loadblock := false;
      exit;
    end;
    if not str2ll (s,ll) then
    begin
      str (linenum,si);
      b := confirm ('Error: wrong coordinate format in line '+si+' of '+
                    filename,false);
      dispose (actblock);
      loadblock := false;
      exit;
    end;
    if minmem > maxavail then
    begin
      b := confirm ('Error: not enough memory available',false);
      loadblock := false;
      exit;
    end;
    new (actcoord);
    actcoord^.name := actblock;
    actblock^.coord := actcoord;
    actcoord^.coordtype := coordtype;
    actcoord^.ll := ll;
    insertblock (actblock,blocklast);
    insertcoord (actcoord,coordlast);
    if not (coordtype in [cline,cpoly]) then
      loadblock := readlnok (s,false)
    else
      while true do
      begin
        repeat
          if not readlnok (s,false) then
          begin
            if coordtype = cline then
              actcoord^.coordtype := clineend
            else
              actcoord^.coordtype := cpolyend;
            loadblock := false;
            exit;
          end;
        until (s <> '') and (copy(s,1,1) <> '.');
        if not str2ll (s,ll) then
        begin
          if coordtype = cline then
            actcoord^.coordtype := clineend
          else
            actcoord^.coordtype := cpolyend;
          loadblock := true;
          exit;
        end;
        if minmem > maxavail then
        begin
          b := confirm ('Error: not enough memory available',false);
          if coordtype = cline then
            actcoord^.coordtype := clineend
          else
            actcoord^.coordtype := cpolyend;
          loadblock := false;
          exit;
        end;
        new (actcoord);
        actcoord^.name := actblock;
        actcoord^.coordtype := coordtype;
        actcoord^.ll := ll;
        insertcoord (actcoord,coordlast);
      end;
  end;

  procedure closefile;
  begin
    {$i-}
    close (f);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t close '+filename,false);
      exit;
    end;
  end;

begin
  loadd := false;
  assign (f,filename);
  {$i-}
  reset (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t open '+filename,false);
    exit;
  end;
  recalc := true;
  loadd := true;
  s := '';
  linenum := 0;
  while true do
  begin
    while (s = '') or (copy(s,1,1) = '.') do
      if not readlnok (s,false) then
      begin
        closefile;
        exit;
      end;
    found := false;
    if filetype in [fcha,fovl,flog] then
      if not loadblock (s,cline) then
      begin
        closefile;
        exit;
      end;
    if filetype in [fcha,fovl] then
      if not loadblock (s,cpoly) then
      begin
        closefile;
        exit;
      end;
    if filetype in [fcha,fovl] then
      if not loadblock (s,ctext) then
      begin
        closefile;
        exit;
      end;
    if filetype in [fcha,fovl,fmem,fwpt] then
      if not loadblock (s,cmark) then
      begin
        closefile;
        exit;
      end;
    if not found then
    begin
      str (linenum,si);
      b := confirm ('Error: unknown block in line '+si+' of '+filename,false);
      exit;
    end;
  end;
  closefile;
end;

function saved (filename : string; filetype : filetypet;
                minmem : longint) : boolean;
  var
    f : text;
    actblock : blockptr;
    actcoord : coordptr;
    s,ss : string;
    b : boolean;

  function writelnok (s : string) : boolean;
    var
      b : boolean;
  begin
    {$i-}
    writeln (f,s);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t write '+filename,false);
      writelnok := false;
    end
    else
      writelnok := true;
  end;

begin
  saved := false;
  if not checkoverwrite (filename) then
    exit;
  assign (f,filename);
  {$i-}
  rewrite (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t open '+filename,false);
    exit;
  end;
  if not writelnok (filecomment) then
    exit;
  actblock := blockfirst;
  while actblock <> nil do
  begin
    if actblock^.filetype = filetype then
    begin
      actcoord := actblock^.coord;
      s := coordtypename[actcoord^.coordtype];
      if actblock^.typec <> ' ' then
        s := s+' '+actblock^.typec;
      if actblock^.scalelimit <> 0 then
      begin
        str (actblock^.scalelimit:6:4,ss);
        s := s+' '+ss;
      end;
      if not writelnok (s) then
        exit;
      if not writelnok (actblock^.name) then
        exit;
      while actcoord^.coordtype in [cline,cpoly] do
      begin
        if not writelnok (ll2str(actcoord^.ll)) then
          exit;
        actcoord := actcoord^.succ;
      end;
      if not writelnok (ll2str(actcoord^.ll)) then
        exit;
    end;
    actblock := actblock^.succ;
  end;
  {$i-}
  close (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t close '+filename,false);
    exit;
  end;
  saved := true;
end;


{ coordinate transform }

function choosegrid (dist : real) : real;
begin
  if dist <= 1 then
    choosegrid := 1
  else if dist <= 10 then
    choosegrid := 10
  else if dist <= 60 then
    choosegrid := 60
  else
    choosegrid := 600;
end;

procedure setchartsize;
  var
    diag1,diag2,b : real;
    uv1,uv2 : uvt;
    xy1,xy2 : xyt;
    ll1,ll2 : llt;
begin
  uv1.u := (xmax+1)*0.5;
  uv1.v := (ymax+1)*0.5;
  uv2xy (uv1,xy1);
  xy2ll (xy1,currentchartcenter);
  uv1.u := (xmax+1)*0.45;
  uv1.v := (ymax+1)*0.45;
  uv2xy (uv1,xy1);
  xy2ll (xy1,ll1);
  uv2.u := (xmax+1)*0.55;
  uv2.v := (ymax+1)*0.55;
  uv2xy (uv2,xy2);
  xy2ll (xy2,ll2);
  calcdistbear (ll1,ll2,diag1,b);
  uv1.u := (xmax+1)*0.55;
  uv1.v := (ymax+1)*0.45;
  uv2xy (uv1,xy1);
  xy2ll (xy1,ll1);
  uv2.u := (xmax+1)*0.45;
  uv2.v := (ymax+1)*0.55;
  uv2xy (uv2,xy2);
  xy2ll (xy2,ll2);
  calcdistbear (ll1,ll2,diag2,b);
  if diag1 > diag2 then
    currentchartsize := diag1
  else
    currentchartsize := diag2;
  currentchartsize := currentchartsize * 10 /
                      sqrt(1+sqr((ymax+1)/(xmax+1)*(yasp/xasp)));
  displaychartsize := currentchartsize;
  defscalelimit := currentchartsize*defscalelimitfact;
  if defscalelimit > 20000 then
    defscalelimit := 0;
  if (griddist <> 0) and (griddist < currentchartsize/maxgridlines) then
    griddist := choosegrid(currentchartsize/maxgridlines);
end;

procedure chartfindab (chartcenter : llt; chartwidth : real);
  var
    llborder : llt;
    xyar : xyart;
    uvar : uvart;
    b : boolean;
begin
  if cos(chartcenter.lat) = 0 then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
    exit;
  end;
  ll2xy (chartcenter,xyar[1]);
  llborder := chartcenter;
  llborder.lon := llborder.lon + chartwidth/60/2/180*pi/cos(chartcenter.lat);
  ll2xy (llborder,xyar[2]);
  xyar[3] := xyar[1];
  xyar[3].y := xyar[3].y - (xyar[2].x-xyar[1].x)*
                           ((ymax+1)/(xmax+1))*(yasp/xasp);
  uvar[1].u := (xmax+1)/2;
  uvar[1].v := (ymax+1)/2;
  uvar[2].u := xmax+1;
  uvar[2].v := (ymax+1)/2;
  uvar[3].u := (xmax+1)/2;
  uvar[3].v := ymax+1;
  if not findab (xyar,uvar) then
  begin
    errorbeep;
    b := confirm ('Error: can''t init transformation',false);
  end
  else
  begin
    bmpdisp := false;
    bmpdisplast := false;
    saveab;
    setchartsize;
    recalc := true;
  end;
end;


{ display text / status }

procedure showtop (left : boolean; s : string);
begin
  if left then
    boxtext (0,0,textwidth(s)+tx1+tx2+5,th+ty1+ty2+5,false,s)
  else
    boxtext (0,0,xmax,th+ty1+ty2+5,true,s);
end;

procedure newactpos;
  var
    s : string;
begin
  uv.u := cx;
  uv.v := cy;
  uv2xy (uv,xy);
  xy2ll (xy,ll);
  s := ll2str (ll);
  newtext ((top1x+top2x) div 2,3,true,s);
end;

procedure initactpos;
  var
    x,y : integer;
begin
  boxtext (top1x,0,top2x-1,th+ty1+ty2+5,true,'');
  newactpos;
end;

procedure showmode (s : string);
begin
  boxtext (top3x,0,top4x-1,th+ty1+ty2+5,true,s);
end;

function brg2str (brg : real) : string;
  var
    s : string;
begin
  str ((brg/pi*180):5:1,s);
  if s[1] = ' ' then
    s[1] := '0';
  if s[2] = ' ' then
    s[2] := '0';
  if s = '360.0' then
    s := '000.0';
  brg2str := s+'';
end;

procedure newgpscur;
  var
    s,ss : string;
    y : integer;
begin
  y := ymax-th-ty1-ty2-2;
  s := brg2str(gpsbrg);
  str (gpsdist:5:3,ss);
  if length(ss) > 5 then
    str (gpsdist:5:2,ss);
  if length(ss) > 5 then
    str (gpsdist:5:1,ss);
  if length(ss) > 5 then
    str (gpsdist:5:0,ss);
  s := s+' '+ss+'nm';
  newtext ((bot3x+bot4x) div 2,y,true,s);
end;

procedure newgpsdata (onlymsg : boolean);
  var
    s,ss : string;
    y : integer;
begin
  boxtext (bot4x,ymax-th-ty1-ty2-5,bot5x-1,ymax,true,gpsmsg);
  if not onlymsg then
  begin
    y := ymax-th-ty1-ty2-2;
    s := ll2str(gpsll);
    newtext ((bot1x+bot2x) div 2,y,true,s);
    s := brg2str(gpscog);
    str (gpssog:4:1,ss);
    s := s+' '+ss+'kn';
    newtext ((bot2x+bot3x) div 2,y,true,s);
    str (currentchartsize:5:3,ss);
    if length(ss) > 5 then
      str (currentchartsize:5:2,ss);
    if length(ss) > 5 then
      str (currentchartsize:5:1,ss);
    if length(ss) > 5 then
      str (currentchartsize:5:0,ss);
    s := ss+'nm';
    newtext ((bot5x+bot6x) div 2,y,true,s);
    newgpscur;
  end;
end;

procedure initgpsdata;
  var
    y1,y2 : integer;
    s : string;
begin
  y2 := ymax;
  y1 := y2-th-ty1-ty2-5;
  s := ll2str(gpsll);
  bot1x := 0;
  bot2x := bot1x+textwidth(s)+tx1+tx2+6;
  s := '000.0 00.0kn';
  bot3x := bot2x+textwidth(s)+tx1+tx2+6;
  s := '000.0 0.000nm';
  bot4x := bot3x+textwidth(s)+tx1+tx2+6;
  s := '0.000nm';
  bot6x := xmax+1;
  bot5x := bot6x-textwidth(s)-tx1-tx2-6;
  boxtext (bot1x,y1,bot2x-1,y2,false,'');
  boxtext (bot2x,y1,bot3x-1,y2,false,'');
  boxtext (bot3x,y1,bot4x-1,y2,false,'');
  boxtext (bot4x,y1,bot5x-1,y2,false,'');
  boxtext (bot5x,y1,bot6x-1,y2,false,'');
  newgpsdata (false);
end;

procedure gpsheadpos (var llh : llt);
  var
    xy : xyt;
    dl,ml,dx,dy : real;
begin
  ll2xy (gpsll,xy);
  dl := gpssog/60*gpsheadtime;
  ml := currentchartsize*gpsheadposfact;
  if dl > ml then
    dl := ml;
  dl := dl/cos(gpsll.lat);
  dx := dl*sin(gpscog);
  dy := dl*cos(gpscog);
  xy.x := xy.x+dx;
  xy.y := xy.y+dy;
  xy2ll (xy,llh);
end;


{ bitmap }

procedure initbitmap (minleft : longint);
begin
  bmpbuf[1] := nil;
  bmpbuf[2] := nil;
  if minleft+longint(((xmax+8) div 8))*(ymax+1) < maxavail then
  begin
    bmpbuf[1] := getbmpbuf (xmax+1,ymax+1);
    if minleft+longint(((xmax+16) div 16))*((ymax+1) div 2) < maxavail then
      bmpbuf[2] := getbmpbuf ((xmax+1) div 2,(ymax+1) div 2);
  end;
end;

procedure calcbitmaps2b (bn : integer; xs,ys : real; var xb,yb : real);
begin
  xb := bmpxoff[bn]+(xs-bmpx0[bn])/bmpxyfact[bn];
  yb := bmpyoff[bn]-(ys-bmpy0[bn])/bmpxyfact[bn];
end;

procedure calcbitmapb2s (bn : integer; xb,yb : real; var xs,ys : real);
begin
  xs := (xb-bmpxoff[bn])*bmpxyfact[bn]+bmpx0[bn];
  ys := -(yb-bmpyoff[bn])*bmpxyfact[bn]+bmpy0[bn];
end;

function loadbitmap (fn : string; xcenter,ycenter,xsize,ysize : integer;
                     zoom,thr,rot : integer; bn : integer) : boolean;
{ OK->true error->false }
{ zoom=-9999 -> fit into size}
  var
    s : string;
    wid,hig,col,t : integer;
    c00 : boolean;
    b : boolean;
begin
  loadbitmap := false;
  if bmpbuf[bn] = nil then
  begin
    b := confirm ('Error: no bitmap buffer memory available',false);
    exit;
  end;
  s := readbmphdr (fn,wid,hig,col,c00);
  if s <> '' then
  begin
    b := confirm ('Error: '+fn+' '+s,false);
    exit;
  end;
  if zoom = -9999 then
  begin
    zoom := 0;
    while wid shr -zoom > xsize do
      dec (zoom);
    while hig shr -zoom > ysize do
      dec (zoom);
  end;
  bmpxyrot[bn] := rot;
  bmpxyfact[bn] := (1 shl (8+zoom))/256;
  bmpxoff[bn] := xcenter-round(xsize/2/bmpxyfact[bn]);
  bmpyoff[bn] := ycenter-round(ysize/2/bmpxyfact[bn]);
  bmpx0[bn] := (xmax+1-xsize) div 2;
  bmpy0[bn] := (ymax+1+ysize) div 2 - 1;
  bmpxs[bn] := xsize;
  bmpys[bn] := ysize;
  if not c00 then
    thr := -thr;
  showtop (false,'loading bitmap ...');
  s := readbmpbuf (fn,bmpxoff[bn],bmpyoff[bn],
                   bmpxs[bn],bmpys[bn],zoom,thr,rot,bmpbuf[bn]);
  if s <> '' then
  begin
    b := confirm ('Error: '+fn+' '+s,false);
    exit;
  end;
  loadbitmap := true;
end;

function showbitmap (bn : integer; bmpcol : integer) : boolean;
{ OK->true error->false }
  var
    s : string;
    b : boolean;
    fgc,bgc : integer;
begin
  showbitmap := false;
  if bmpbuf[bn] = nil then
    exit;
  if not reverse then
  begin
    bgc := 0;
    fgc := 0+bmpcol;
  end
  else
  begin
    bgc := getmaxcolor;
    fgc := bgc-bmpcol;
  end;
  newscreen;
  xhidecursor;
  s := showbmpbuf (bmpbuf[bn],bmpx0[bn],bmpy0[bn],bmpxs[bn],bmpys[bn],
                   bgc,fgc);
  xshowcursor;
  if s <> '' then
  begin
    b := confirm ('Error: '+s,false);
    exit;
  end;
  showbitmap := true;
end;

function savebitmap (fn : string) : boolean;
{ OK->true error->false }
  var
    s : string;
    b : boolean;
begin
  savebitmap := false;
  xhidecursor;
  s := writebmp (fn,0,ymax,xmax+1,ymax+1,0);
  xshowcursor;
  if s <> '' then
  begin
    b := confirm ('Error: '+fn+' '+s,false);
    exit;
  end;
  savebitmap := true;
end;

function loadbmc (bmcfn : string; var bmpfn : string; var thr,rot : integer;
                  var llar : llart; var uvar : uvart) : boolean;
{ OK->true error->false }
  var
    f : text;
    s,s1,s2 : string;
    b : boolean;
    i,p : integer;
    er : integer;

  function readline (var s : string) : boolean;
  begin
    readline := false;
    {$i-}
    readln (f,s);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t read '+bmcfn,false);
      close (f);
      exit;
    end;
    readline := true;
  end;

begin
  loadbmc := false;
  assign (f,bmcfn);
  {$i-}
  reset (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t open '+bmcfn,false);
    exit;
  end;
  if not readline (s) then
    exit;
  if s <> 'bitmap chart' then
  begin
    b := confirm ('Error: no bitmap chart header',false);
    close (f);
    exit;
  end;
  if not readline (bmpfn) then
    exit;
  if s = '' then
  begin
    b := confirm ('Error: no bitmap filename',false);
    close (f);
    exit;
  end;
  if not readline (s) then
    exit;
  while (length(s) > 0) and (s[1] = ' ') do
    s := copy(s,2,length(s)-1);
  p := pos(' ',s);
  if p = 0 then
  begin
    s1 := s;
    s2 := '0';
  end
  else
  begin
    s1 := copy(s,1,p-1);
    s2 := copy(s,p+1,length(s)-p);
  end;
  val (s1,thr,er);
  if er = 0 then
    val (s2,rot,er);
  if er <> 0 then
  begin
    b := confirm ('Error: wrong threshold/rotation format',false);
    close (f);
    exit;
  end;
  for i := 1 to 3 do
  begin
    if not readline (s) then
      exit;
    if not str2ll (s,llar[i]) then
    begin
      b := confirm ('Error: wrong coordinate format',false);
      close (f);
      exit;
    end;
  end;
  for i := 1 to 3 do
  begin
    if not readline (s) then
      exit;
    while (length(s) > 0) and (s[1] = ' ') do
      s := copy(s,2,length(s)-1);
    p := pos(' ',s);
    if p = 0 then
      er := 1
    else
    begin
      s1 := copy(s,1,p-1);
      s2 := copy(s,p+1,length(s)-p);
    end;
    if er = 0 then
      val (s1,uvar[i].u,er);
    if er = 0 then
      val (s2,uvar[i].v,er);
    if er <> 0 then
    begin
      b := confirm ('Error: wrong bmp coordinate format',false);
      close (f);
      exit;
    end;
  end;
  close (f);
  loadbmc := true;
end;

function savebmc (bmcfn : string; bmpfn : string; thr,rot : integer;
                  llar : llart; uvar : uvart) : boolean;
{ OK->true error->false }
  var
    f : text;
    s,s1,s2 : string;
    b : boolean;
    i : integer;

  function writeline (s : string) : boolean;
  begin
    writeline := false;
    {$i-}
    writeln (f,s);
    {$i+}
    if ioresult <> 0 then
    begin
      b := confirm ('Error: can''t write '+bmcfn,false);
      close (f);
      exit;
    end;
    writeline := true;
  end;

begin
  savebmc := false;
  assign (f,bmcfn);
  {$i-}
  rewrite (f);
  {$i+}
  if ioresult <> 0 then
  begin
    b := confirm ('Error: can''t open '+bmcfn,false);
    exit;
  end;
  if not writeline ('bitmap chart') then
    exit;
  if not writeline (bmpfn) then
    exit;
  str (thr,s1);
  str (rot,s2);
  if not writeline (s1+' '+s2) then
    exit;
  for i := 1 to 3 do
    if not writeline (ll2str(llar[i])) then
      exit;
  for i := 1 to 3 do
  begin
    str (uvar[i].u:8:1,s1);
    str (uvar[i].v:8:1,s2);
    if not writeline (s1+' '+s2) then
      exit;
  end;
  close (f);
  savebmc := true;
end;


begin
end.

{ end unit }
