unit grafbmp;

{
************************************************************

grafbmp.pas - .bmp bitmap file read/write

(c) Heiko Purnhagen 1993-1997
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/purnhage/

960723   born
960725   ...
970803   error if zoom > 2
971207   bugfixes (shifting bmp 1 pel down)
971207   bmpbuf
971225   ...
971229   blockread/blockwrite
971231   new interface
980421   fixed pba[..] := 0 index range
980615   fixed showbmpbuf for b/w

************************************************************
}

interface

uses
  graph;

type
  { internal b/w bitmap format }
  bmpbuft = record
              wid : word;
              hig : word;
              byteline : word;
              size : word;
              buf : pointer;
            end;
  bmpbufp = ^bmpbuft;

function readbmphdr (filename : string;
                     var wid,hig,col : integer; var c00 : boolean) : string;
{ ok->'' error->'<error msg>' }

function readbmp (filename : string; x0,y0 : integer;
                  inv : boolean) : string;
{ ok->'' error->'<error msg>' }
{ x0,y0=lower left corner on screen }
{ inv: true -> inverse colors }
{ max bmp size: 640*480   b/w or col16 }

function writebmp (filename : string; x0,y0,xs,ys : integer;
                   thr : integer) : string;
{ ok->'' error->'<error msg>' }
{ x0,y0=lower left corner on screen }
{ xs,ys=size on screen }
{ thr=0 -> original b/w or col16   thr>0 -> b/w using threshold }
{ max bmp size: 640*480   b/w or col16 }

function getbmpbuf (xs,ys : integer) : bmpbufp;
{ error->nil }
{ max size: 640*480 }

function readbmpbuf (filename : string; xoff,yoff : integer;
                     xs,ys,zoom,thr,rot : integer;
                     bmpbuf : bmpbufp) : string;
{ ok->'' error->'<error msg>' }
{ xoff,yoff=bmp offset (lower left corner) }
{ xs,ys=size }
{ zoom=-4->*1/16 zoom=0->*1 zoom=2->*4   thr=1->norm thr=-1->inv }
{ rot: 0=0 1=90 2=180 3=270 deg cw }
{ max bmp size: 8000*8000   max zoom: 2 (400%) }

function showbmpbuf (bmpbuf : bmpbufp;
                     x0,y0,xs,ys,bgc,fgc : integer) : string;
{ ok->'' error->'<error msg>' }
{ x0,y0=lower left corner on screen }
{ xs,ys=size on screen }
{ cga/att400: b/w    fgc=1: normal  fgc=0: inverse }
{ ega/vga:    col16  background: bgc  foreground: fgc }

procedure freebmpbuf (bmpbuf : bmpbufp);


implementation

const
  minmem = 1000;
  fbufsize = 1024;   { size of col256 palette }

type
  pbat = array [0..38399] of byte;   { size of 640*480 b/w }
  fbuft = array [0..fbufsize-1] of byte;

  { getimage/putimage data format }
  pbit = record
           sx1 : word;   { width: sx1+1 pel }
           sy1 : word;   { height: sy1+1 pel }
           pix : array [0..3199] of byte;   { 16col 640x10 }
           { data: line by line, top line first, msb = left pixel }
           {       16col: bright-line, red-line, ... }
           {   cga     b/w:   8 pel / 1 byte }
           {   cga     4col:  4 pel / 1 byte (not supported here) }
           {   ega/vga 16col: 8 pel / 4 byte (bright,red,green,blue) }
           dummy : integer;   { to match imagesize() ... }
         end;

  { bmp file data format }
  { see also: http://www.ice-digga.com/programming/bmp.html }
  { bmp header }
  bmphdrt = record
              magic :  word;      { magic number ($4d42 = 'BM') }
              totlen : longint;   { total size [byte] }
              rsrvd1 : word;
              rsrvd2 : word;
              offdat : longint;   { offset for bmp data [byte] }
              hdrlen : longint;   { info header length [byte] ($0028) }
              width :  longint;   { width [pel] }
              height : longint;   { height [pel] }
              collay : word;      { color layers ($01 if rgb palette) }
              colbit : word;      { color bits  (numcol = 1 shl colbit) }
              compre : longint;   { compression mode ($00 = no compr.) }
              datlen : longint;   { data length [byte] }
              xppm :   longint;   { horizontal pel per meter ($0b13 = 72dpi) }
              yppm :   longint;   { vertical pel per meter ($0b13 = 72dpi) }
              clrusd : longint;   { num of palette entries used ($00) }
              clrimp : longint;   { num of palette entries important ($00) }
            end;
  { bmp rgb palette   array[0..numcol-1] of bmppalt }
  bmppalt = array [0..3] of byte;   { blue, green, red, dummy }
                                    { $00 $00 $00 $00 .. $ff $ff $ff $00 }
  { bmp data (here: collay=1, colbit=1..8) }
  {   bmppict : array [0..height-1] of bmpline;   1st line: bottom }
  {     bmpline : array [0..byteline-1] of byte;   1st byte msb(s): left pel }
  {               byteline := ((width*colbit+31) div 32)*4 }

const
  { bmp default header }
  bmphdrbw :  bmphdrt = (magic:  $4d42;
                         totlen: $00000000;
                         rsrvd1: $0000;
                         rsrvd2: $0000;
                         offdat: $0000003e;
                         hdrlen: $00000028;
                         width:  $00000000;
                         height: $00000000;
                         collay: $0000;
                         colbit: $0000;
                         compre: $00000000;
                         datlen: $00000000;
                         xppm:   $00000b13;
                         yppm:   $00000b13;
                         clrusd: $00000000;
                         clrimp: $00000000);
  bmphdrc16 : bmphdrt = (magic:  $4d42;
                         totlen: $00000000;
                         rsrvd1: $0000;
                         rsrvd2: $0000;
                         offdat: $00000076;
                         hdrlen: $00000028;
                         width:  $00000000;
                         height: $00000000;
                         collay: $0000;
                         colbit: $0000;
                         compre: $00000000;
                         datlen: $00000000;
                         xppm:   $00000b13;
                         yppm:   $00000b13;
                         clrusd: $00000000;
                         clrimp: $00000000);
  { bmp rgb palette: b/w }
  bwpal :  array [0..1] of bmppalt = (($00, $00, $00, $00),
                                      ($ff, $ff, $ff, $00));
  { bmp rgb palette: col16 }
  c16pal : array [0..15] of bmppalt = (($00, $00, $00, $00),
                                       ($00, $00, $80, $00),
                                       ($00, $80, $00, $00),
                                       ($00, $80, $80, $00),
                                       ($80, $00, $00, $00),
                                       ($80, $00, $80, $00),
                                       ($80, $80, $00, $00),
                                       ($c0, $c0, $c0, $00),
                                       ($80, $80, $80, $00),
                                       ($00, $00, $ff, $00),
                                       ($00, $ff, $00, $00),
                                       ($00, $ff, $ff, $00),
                                       ($ff, $00, $00, $00),
                                       ($ff, $00, $ff, $00),
                                       ($ff, $ff, $00, $00),
                                       ($ff, $ff, $ff, $00));

var
  pbi : pbit;
  fbuf : fbuft;
  bmphdr : bmphdrt;


function readbmphdrf (var f : file; var wid,hig,colbit,col : integer;
                      var c00 : boolean) : string;
{ ok->'' error->'<error msg>' }
  var
    num,res : word;
begin
  wid := 0;
  hig := 0;
  colbit := 0;
  col := 0;
  c00 := true;
  { bmp header }
  {$i-}
  blockread (f,bmphdr,sizeof(bmphdrt),res);
  {$i+}
  if (ioresult <> 0) or (res <> sizeof(bmphdrt)) then
  begin
    readbmphdrf := 'file too short';
    exit;
  end;
  if bmphdr.magic <> $4d42 then
  begin
    readbmphdrf := 'no bmp header';
    exit;
  end;
  wid := bmphdr.width;
  hig := bmphdr.height;
  if bmphdr.collay <> $0001 then
  begin
    readbmphdrf := 'more than 1 layer';
    exit;
  end;
  if bmphdr.colbit > 8 then
  begin
    readbmphdrf := 'more than 256 colors';
    exit;
  end;
  colbit := bmphdr.colbit;
  col := 1 shl bmphdr.colbit;
  { rgb palette }
  num := 4 * (1 shl bmphdr.colbit);
  {$i-}
  blockread (f,fbuf,num,res);
  {$i+}
  if (ioresult <> 0) or (res <> num) then
  begin
    readbmphdrf := 'file too short';
    exit;
  end;
  c00 := (fbuf[0] = 0) and (fbuf[1] = 0) and (fbuf[2] = 0);
  readbmphdrf := '';
end;


function readbmphdr (filename : string;
                     var wid,hig,col : integer; var c00 : boolean) : string;
{ ok->'' error->'<error msg>' }
  var
    f : file;
    colbit : integer;
begin
  assign (f,filename);
  {$i-}
  reset (f,1);
  {$i+}
  if ioresult <> 0 then
  begin
    readbmphdr := 'can''t open file';
    exit;
  end;
  readbmphdr := readbmphdrf(f,wid,hig,colbit,col,c00);
  close (f);
end;

function readbmp (filename : string; x0,y0 : integer;
                  inv : boolean) : string;
{ ok->'' error->'<error msg>' }
{ x0,y0=lower left corner on screen }
{ inv: true -> inverse colors }
{ max bmp size: 640*480   b/w or col16 }
  var
    f : file;
    res : word;
    s : string;
    wid,hig,colbit,col : integer;
    c00 : boolean;
    b,invmask : byte;
    mc,byteline,byte1 : integer;
    y,yd,i,ii,i0,i1,j,sr : integer;
begin
  assign (f,filename);
  {$i-}
  reset (f,1);
  {$i+}
  if ioresult <> 0 then
  begin
    readbmp := 'can''t open file';
    exit;
  end;
  s := readbmphdrf(f,wid,hig,colbit,col,c00);
  if s <> '' then
  begin
    readbmp := s;
    close (f);
    exit;
  end;
  if (wid > 640) or (hig > 480) then
  begin
    readbmp := 'bmp too large';
    close (f);
    exit;
  end;
  if (x0 < 0) or (y0-hig+1 < 0) or
     (x0+wid-1 > getmaxx) or (y0 > getmaxy) then
  begin
    readbmp := 'bmp exceeds screen';
    close (f);
    exit;
  end;
  mc := getmaxcolor+1;
  if (mc <> 2) and (mc <> 16) then
  begin
    readbmp := 'screen not b/w or col16';
    close (f);
    exit;
  end;
  if (col <> 2) and (col <> 16) then
  begin
    readbmp := 'bmp not b/w or col16';
    close (f);
    exit;
  end;
  if (mc = 2) and (col = 16) then
  begin
    readbmp := 'bmp col16 but screen b/w';
    close (f);
    exit;
  end;
  if imagesize(0,0,639,9) > sizeof(pbit) then
  begin
    readbmp := 'imagesize error';
    close (f);
    exit;
  end;

  byteline := ((wid*colbit+31) div 32)*4;
  byte1 := ((wid*colbit+7) div 8)-1;
  invmask := $00;
  if inv xor (not c00) then
    invmask := $ff;
  ii := ((wid-1) div 8) + 1;
  y := 0;
  repeat
    yd := hig-y;
    if yd > 10 then
      yd := 10;
    pbi.sx1 := wid-1;
    pbi.sy1 := yd-1;
    for j := 0 to yd-1 do
    begin
      {$i-}
      blockread (f,fbuf,byteline,res);
      {$i+}
      if (ioresult <> 0) or (res <> byteline) then
      begin
        readbmp := 'file too short';
        close (f);
        exit;
      end;
      if col = 16 then
      begin
        { col16 }
        i0 := (yd-1-j)*4*ii;
        for i := 0 to 4*ii-1 do
          pbi.pix[i0+i] := 0;
        for i := 0 to byte1 do
        begin
          sr := 2*(i mod 4);
          b := fbuf[i] xor invmask;
          i1 := i0+(i div 4);
          pbi.pix[i1+0*ii] := pbi.pix[i1+0*ii] or (((b shl 0) and $80) shr sr);
          pbi.pix[i1+1*ii] := pbi.pix[i1+1*ii] or (((b shl 3) and $80) shr sr);
          pbi.pix[i1+2*ii] := pbi.pix[i1+2*ii] or (((b shl 2) and $80) shr sr);
          pbi.pix[i1+3*ii] := pbi.pix[i1+3*ii] or (((b shl 1) and $80) shr sr);
          sr := sr+1;
          b := b shl 4;
          pbi.pix[i1+0*ii] := pbi.pix[i1+0*ii] or (((b shl 0) and $80) shr sr);
          pbi.pix[i1+1*ii] := pbi.pix[i1+1*ii] or (((b shl 3) and $80) shr sr);
          pbi.pix[i1+2*ii] := pbi.pix[i1+2*ii] or (((b shl 2) and $80) shr sr);
          pbi.pix[i1+3*ii] := pbi.pix[i1+3*ii] or (((b shl 1) and $80) shr sr);
        end;
      end
      else if mc = 16 then
      begin
        { b/w -> col16 }
        i0 := (yd-1-j)*4*ii;
        for i := 0 to byte1 do
        begin
          pbi.pix[i0+0*ii+i] := fbuf[i] xor invmask;
          pbi.pix[i0+1*ii+i] := fbuf[i] xor invmask;
          pbi.pix[i0+2*ii+i] := fbuf[i] xor invmask;
          pbi.pix[i0+3*ii+i] := fbuf[i] xor invmask;
        end;
      end
      else
      begin
        { b/w }
        i0 := (yd-1-j)*ii;
        for i := 0 to byte1 do
          pbi.pix[i0+i] := fbuf[i] xor invmask;
      end;
    end;
    putimage (x0,y0-y-yd+1,pbi,normalput);
    y := y+yd;
  until y = hig;
  readbmp := '';
  close (f);
end;


function writebmp (filename : string; x0,y0,xs,ys : integer;
                   thr : integer) : string;
{ ok->'' error->'<error msg>' }
{ x0,y0=lower left corner on screen }
{ xs,ys=size on screen }
{ thr=0 -> original b/w or col16   thr>0 -> b/w using threshold }
{ max bmp size: 640*480   b/w or col16 }
  var
    f : file;
    res : word;
    wid,hig,colbit : integer;
    b,d0,d1,d2,d3,t0,t1,t2,t3 : byte;
    mc,byteline,byte1 : integer;
    y,yd,i,ii,i0,i1,j,sr,k,kk : integer;
begin
  wid := xs;
  hig := ys;
  if (wid > 640) or (hig > 480) then
  begin
    writebmp := 'bmp too large';
    exit;
  end;
  if (x0 < 0) or (y0-hig+1 < 0) or
     (x0+wid-1 > getmaxx) or (y0 > getmaxy) then
  begin
    writebmp := 'bmp exceeds screen';
    exit;
  end;
  mc := getmaxcolor+1;
  if (mc <> 2) and (mc <> 16) then
  begin
    writebmp := 'screen not b/w or col16';
    exit;
  end;
  colbit := 1;
  if (mc = 16) and (thr = 0) then
    colbit := 4;
  if imagesize(0,0,639,9) > sizeof(pbit) then
  begin
    writebmp := 'imagesize error';
    exit;
  end;

  assign (f,filename);
  {$i-}
  rewrite (f,1);
  {$i+}
  if ioresult <> 0 then
  begin
    writebmp := 'can''t open file';
    exit;
  end;
  { bmp header }
  if colbit = 1 then
    bmphdr := bmphdrbw
  else
    bmphdr := bmphdrc16;
  bmphdr.width := wid;
  bmphdr.height := hig;
  bmphdr.collay := 1;
  bmphdr.colbit := colbit;
  byteline := ((wid*colbit+31) div 32)*4;
  bmphdr.datlen := longint(byteline)*hig;
  bmphdr.totlen := bmphdr.datlen+sizeof(bmphdr)+(4 shl colbit);
  {$i-}
  blockwrite (f,bmphdr,sizeof(bmphdrt),res);
  {$i+}
  if (ioresult <> 0) or (res <> sizeof(bmphdrt)) then
  begin
    writebmp := 'error writing file';
    close (f);
    exit;
  end;
  { rgb palette }
  if colbit = 1 then
  begin
    {$i-}
    blockwrite (f,bwpal,sizeof(bwpal),res);
    {$i+}
    if (ioresult <> 0) or (res <> sizeof(bwpal)) then
    begin
      writebmp := 'error writing file';
      close (f);
      exit;
    end;
  end
  else
  begin
    {$i-}
    blockwrite (f,c16pal,sizeof(c16pal),res);
    {$i+}
    if (ioresult <> 0) or (res <> sizeof(c16pal)) then
    begin
      writebmp := 'error writing file';
      close (f);
      exit;
    end;
  end;

  byte1 := ((wid*colbit+7) div 8)-1;
  if thr > 0 then
  begin
    t0 := -(((thr-1) shr 0) and $01);
    t1 := -(((thr-1) shr 1) and $01);
    t2 := -(((thr-1) shr 2) and $01);
    t3 := -(((thr-1) shr 3) and $01);
  end;
  ii := ((wid-1) div 8) + 1;
  y := 0;
  repeat
    yd := hig-y;
    if yd > 10 then
      yd := 10;
    getimage (x0,y0-y-yd+1,x0+wid-1,y0-y,pbi);
    for j := 0 to yd-1 do
    begin
      if colbit = 4 then
      begin
        { col16 }
        i0 := (yd-1-j)*4*ii;
        for i := 0 to byte1 do
        begin
          b := 0;
          sr := 7-2*(i mod 4);
          i1 := i0+(i div 4);
          b := b or (((pbi.pix[i1+0*ii] shr sr) and $01) shl 3);
          b := b or (((pbi.pix[i1+1*ii] shr sr) and $01) shl 0);
          b := b or (((pbi.pix[i1+2*ii] shr sr) and $01) shl 1);
          b := b or (((pbi.pix[i1+3*ii] shr sr) and $01) shl 2);
          sr := sr-1;
          b := b shl 4;
          b := b or (((pbi.pix[i1+0*ii] shr sr) and $01) shl 3);
          b := b or (((pbi.pix[i1+1*ii] shr sr) and $01) shl 0);
          b := b or (((pbi.pix[i1+2*ii] shr sr) and $01) shl 1);
          b := b or (((pbi.pix[i1+3*ii] shr sr) and $01) shl 2);
          fbuf[i] := b;
        end;
      end
      else if mc = 16 then
      begin
        { col16 -> b/w }
        i0 := (yd-1-j)*4*ii;
        for i := 0 to byte1 do
        begin
          d3 := pbi.pix[i0+0*ii+i];
          d0 := pbi.pix[i0+1*ii+i];
          d1 := pbi.pix[i0+2*ii+i];
          d2 := pbi.pix[i0+3*ii+i];
          fbuf[i] := d3 and not t3 or
                     (d3 xor not t3) and (d2 and not t2 or
                     (d2 xor not t2) and (d1 and not t1 or
                     (d1 xor not t1) and (d0 and not t0)));
        end;
      end
      else
      begin
        { b/w }
        i0 := (yd-1-j)*ii;
        for i := 0 to byte1 do
          fbuf[i] := pbi.pix[i0+i];
      end;
      {$i-}
      blockwrite (f,fbuf,byteline,res);
      {$i+}
      if (ioresult <> 0) or (res <> byteline) then
      begin
        writebmp := 'error writing file';
        close (f);
        exit;
      end;
    end;
    y := y+yd;
  until y = hig;
  writebmp := '';
  close (f);
end;


{ bmpbuf functions }

function getbmpbuf (xs,ys : integer) : bmpbufp;
{ error->nil }
{ max size: 640*480 }
  var
    bmpbuf : bmpbufp;
begin
  getbmpbuf := nil;
  if (xs <= 0) or (ys <= 0) then
    exit;
  if (xs > 640) or (ys > 480) then
    exit;
  if minmem > maxavail then
    exit;
  new (bmpbuf);
  bmpbuf^.wid := xs;
  bmpbuf^.hig := ys;
  bmpbuf^.byteline := (xs+7) div 8;
  bmpbuf^.size := bmpbuf^.byteline * ys;
  if (bmpbuf^.size > sizeof(pbat)) or (minmem+bmpbuf^.size > maxavail) then
  begin
    dispose (bmpbuf);
    exit;
  end;
  getmem (bmpbuf^.buf,bmpbuf^.size);
  getbmpbuf := bmpbuf;
end;

function readbmpbuf (filename : string; xoff,yoff : integer;
                     xs,ys,zoom,thr,rot : integer;
                     bmpbuf : bmpbufp) : string;
{ ok->'' error->'<error msg>' }
{ xoff,yoff=bmp offset (lower left corner) }
{ xs,ys=size }
{ zoom=-4->*1/16 zoom=0->*1 zoom=2->*4   thr=1->norm thr=-1->inv }
{ rot: 0=0 1=90 2=180 3=270 deg cw }
{ max bmp size: 8000*8000   max zoom: 2 (400%) }
  var
    f : file;
    s : string;
    c00 : boolean;
    w : word;
    b : byte;
    i : integer;
    wid,hig,col,colbit,colmask,colshift : integer;
    byteline,pelbyte,bytedisp : integer;
    bytepre,bytepost : integer;
    r,c,cc,cd,x,y,xd,xi,yi,pelsize,byted : integer;
    pel,pelinv : boolean;
    t,xx,yy,xx0,yy0 : integer;
    pba : ^pbat;
    pbl : word;
    bb : ^byte;
    fbnum,fbact,fbrem : word;

  function getfbuf : boolean;
  { error->true }
  begin
    getfbuf := true;
    while fbact >= fbnum do
    begin
      fbact := fbact - fbnum;
      {$i-}
      blockread (f,fbuf,fbufsize,fbnum);
      {$i+}
      if (ioresult <> 0) then
        exit;
      if (fbnum < fbufsize) and (fbact >= fbnum) then
        exit;
    end;
    getfbuf := false;
  end;

begin
  if bmpbuf = nil then
  begin
    readbmpbuf := 'bmpbuf nil';
    exit;
  end;
  if (xs > bmpbuf^.wid) or (ys > bmpbuf^.hig) then
  begin
    readbmpbuf := 'bmpbuf too small';
    exit;
  end;
  if (xs <= 0) or (ys <= 0) then
  begin
    readbmpbuf := '';
    exit;
  end;
  if zoom > 2 then
  begin
    readbmpbuf := 'zoom > 2';
    exit;
  end;
  if (rot < 0) or (rot > 3) then
  begin
    readbmpbuf := 'rot not 0,1,2,3';
    close (f);
    exit;
  end;
  assign (f,filename);
  {$i-}
  reset (f,1);
  {$i+}
  if ioresult <> 0 then
  begin
    readbmpbuf := 'can''t open file';
    exit;
  end;
  s := readbmphdrf(f,wid,hig,colbit,col,c00);
  if s <> '' then
  begin
    readbmpbuf := s;
    close (f);
    exit;
  end;
  if (wid > 8000) or (hig > 8000) then
  begin
    readbmpbuf := 'bmp too large';
    exit;
  end;

  colmask := col-1;
  if zoom > 0 then
    pelsize := (1 shl zoom)-1
  else
    pelsize := 0;
  case rot of
    0 : begin
          xx0 := 0;
          yy0 := ys-1-pelsize;
        end;
    1 : begin
          xx0 := 0;
          yy0 := 0;
        end;
    2 : begin
          xx0 := xs-1-pelsize;
          yy0 := 0;
        end;
    3 : begin
          xx0 := xs-1-pelsize;
          yy0 := ys-1-pelsize;
        end;
  end;
  if (rot=1) or (rot=3) then
  begin
    t := xs;
    xs := ys;
    ys := t;
  end;
  pba := bmpbuf^.buf;
  pbl := bmpbuf^.byteline;
  byteline := ((wid*colbit+31) div 32)*4;
  pelinv := false;
  if thr < 0 then
  begin
    thr := -thr;
    pelinv := true;
  end;
  colmask := colmask shl (8-colbit);
  colshift := colbit;
  cd := 1;
  xd := 1;
  bytedisp := 1;
  if zoom < 0 then
  begin
    cd := 1 shl -zoom;
    colshift := colbit shl -zoom;
    bytedisp := (colshift+7) div 8
  end;
  if zoom > 0 then
    xd := 1 shl zoom;
  pelbyte := 8 div colbit;
  bytepre := xoff div pelbyte;
  bytepost := byteline-1-
              ((((xs shl (2-zoom)) shr 2)+xoff) div pelbyte);
  if bytepre < 0 then
    bytepre := 0;
  if bytepost < 0 then
    bytepost := 0;
  thr := thr shl (8-colbit);
  byted := ((pelbyte div cd)-1)*xd;
  for w := 0 to bmpbuf^.size-1 do
    pba^[w] := 0;
  fbact := 0;
  fbnum := 0;

  for r := 0-yoff to hig-1-yoff do
  begin
    if r < 0 then
      y := -1
    else
      y := (r shl 2) shr (2-zoom);
    if (y < 0) or (y+pelsize >= ys) or
       ((zoom < 0) and (r mod (1 shl -zoom) <> 0)) then
    begin
      if (fbact >= fbnum) and getfbuf then
      begin
        readbmpbuf := 'file too short';
        close (f);
        exit;
      end;
      fbact := fbact+byteline;
    end
    else
    begin
      fbrem := byteline-bytepre;
      fbact := fbact+bytepre;
      c := bytepre*pelbyte-xoff;
      case rot of
        0 : yy := yy0-y;
        1 : xx := xx0+y;
        2 : yy := yy0+y;
        3 : xx := xx0-y;
      end;
      for i := 1 to (byteline-bytepre-bytepost) div bytedisp do
      begin
        if (fbact >= fbnum) and getfbuf then
        begin
          readbmpbuf := 'file too short';
          close (f);
          exit;
        end;
        b := fbuf[fbact];
        fbact := fbact+bytedisp;
        fbrem := fbrem-bytedisp;
        if c < 0 then
          x := (((c+4096) shl 2) shr (2-zoom)) - (16384 shr (2-zoom))
        else
          x := (c shl 2) shr (2-zoom);
        cc := c+pelbyte;
        if cc > wid-xoff then
          cc := wid-xoff;
        if (x < 0) or (x+byted+pelsize > xs-1) then
        begin
          while c < cc do
          begin
            if (x >= 0) and (x+pelsize <= xs-1) then
            begin
              if ((b and colmask) < thr) = pelinv then
              begin
                case rot of
                  0 : xx := xx0+x;
                  1 : yy := yy0+x;
                  2 : xx := xx0-x;
                  3 : yy := yy0-x;
                end;
                if pelsize = 0 then
                begin
                  bb := addr(pba^[word(yy)*pbl+(xx shr 3)]);
                  bb^ := bb^ or (128 shr (xx and 7));
                end
                else
                  for xi := 0 to pelsize do
                    for yi := 0 to pelsize do
                    begin
                      bb := addr(pba^[word(yy+yi)*pbl+((xx+xi) shr 3)]);
                      bb^ := bb^ or (128 shr ((xx+xi) and 7));
                    end
              end;
            end;
            b := b shl colshift;
            c := c+cd;
            x := x+xd;
          end;
        end
        else
        begin
          if pelsize = 0 then
          begin
            while c < cc do
            begin
              if ((b and colmask) < thr) = pelinv then
              begin
                case rot of
                  0 : xx := xx0+x;
                  1 : yy := yy0+x;
                  2 : xx := xx0-x;
                  3 : yy := yy0-x;
                end;
                bb := addr(pba^[word(yy)*pbl+(xx shr 3)]);
                bb^ := bb^ or (128 shr (xx and 7));
              end;
              b := b shl colshift;
              c := c+cd;
              x := x+xd;
            end;
          end
          else
          begin
            while c < cc do
            begin
              if ((b and colmask) < thr) = pelinv then
              begin
                case rot of
                  0 : xx := xx0+x;
                  1 : yy := yy0+x;
                  2 : xx := xx0-x;
                  3 : yy := yy0-x;
                end;
                for xi := 0 to pelsize do
                  for yi := 0 to pelsize do
                  begin
                    bb := addr(pba^[word(yy+yi)*pbl+((xx+xi) shr 3)]);
                    bb^ := bb^ or (128 shr ((xx+xi) and 7));
                  end
              end;
              b := b shl colshift;
              c := c+cd;
              x := x+xd;
            end;
          end;
        end;
      end;   { for i := ... }
      fbact := fbact+fbrem;
    end;
  end;   { for r := ... }

  close (f);
  readbmpbuf := '';
end;


function showbmpbuf (bmpbuf : bmpbufp;
                     x0,y0,xs,ys,bgc,fgc : integer) : string;
{ ok->'' error->'<error msg>' }
{ x0,y0=lower left corner on screen }
{ xs,ys=size on screen }
{ cga/att400: b/w    fgc=1: normal  fgc=0: inverse }
{ ega/vga:    col16  background: bgc  foreground: fgc }
  var
    mc : integer;
    y,yd : word;
    i,ii,ii1,i0 : word;
    j,j0 : word;
    t,m : byte;
    bar,xar : array [0..3] of byte;
    pba : ^pbat;
    pbl : word;
begin
  if bmpbuf = nil then
  begin
    showbmpbuf := 'bmpbuf = nil';
    exit;
  end;
  if (xs <= 0) or (ys <= 0) then
  begin
    showbmpbuf := '';
    exit;
  end;
  if (xs > bmpbuf^.wid) or (ys > bmpbuf^.hig) or (xs > 640) then
  begin
    showbmpbuf := 'xs,ys too large';
    exit;
  end;
  if (x0 < 0) or (y0-ys+1 < 0) or
     (x0+xs-1 > getmaxx) or (y0 > getmaxy) then
  begin
    showbmpbuf := 'bmpbuf exceeds screen';
    exit;
  end;

  mc := getmaxcolor+1;
  if mc = 16 then
  begin
    { col16 }
    bar[0] := ((bgc shr 3) and 1)*255;
    bar[1] := ((bgc shr 2) and 1)*255;
    bar[2] := ((bgc shr 1) and 1)*255;
    bar[3] := ((bgc shr 0) and 1)*255;
    xar[0] := bar[0] xor ((fgc shr 3) and 1)*255;
    xar[1] := bar[1] xor ((fgc shr 2) and 1)*255;
    xar[2] := bar[2] xor ((fgc shr 1) and 1)*255;
    xar[3] := bar[3] xor ((fgc shr 0) and 1)*255;
  end
  else if mc = 2 then
  begin
    { b/w }
    m := 255;
    if fgc <> 0 then
      m := 0;
  end
  else
  begin
    showbmpbuf := 'not b/w or col16';
    exit;
  end;
  if imagesize(0,0,639,9) > sizeof(pbit) then
  begin
    showbmpbuf := 'imagesize error';
    exit;
  end;
  ii1 := (xs-1) div 8;
  ii := ii1+1;
  pba := bmpbuf^.buf;
  pbl := bmpbuf^.byteline;
  y := 0;
  repeat
    yd := ys-y;
    if yd > 10 then
      yd := 10;
    pbi.sx1 := xs-1;
    pbi.sy1 := yd-1;
    if mc = 16 then
    begin
      { col16 }
      for j := 0 to yd-1 do
      begin
        j0 := word(y+j)*pbl;
        i0 := j*4*ii;
        for i := 0 to ii1 do
        begin
          t := pba^[j0+i];
          pbi.pix[i0     +i] := bar[0] xor (t and xar[0]);
          pbi.pix[i0+  ii+i] := bar[1] xor (t and xar[1]);
          pbi.pix[i0+2*ii+i] := bar[2] xor (t and xar[2]);
          pbi.pix[i0+3*ii+i] := bar[3] xor (t and xar[3]);
        end;
      end;
    end
    else
    begin
      { b/w }
      for j := 0 to yd-1 do
      begin
        j0 := word(y+j)*pbl;
        i0 := j*ii;
        for i := 0 to ii1 do
          pbi.pix[i0+i] := pba^[j0+i] xor m;
      end;
    end;
    putimage(x0,y0-ys+1+y,pbi,normalput);
    y := y+yd;
  until y = ys;
  showbmpbuf := '';
end;

procedure freebmpbuf (bmpbuf : bmpbufp);
begin
  if bmpbuf = nil then
    exit;
  freemem (bmpbuf^.buf,bmpbuf^.size);
  release (bmpbuf);
end;


{ init unit }
begin
end.

{ end unit }
