unit mousedrv;

{
************************************************************

mousedrv.pas - unit for mouse

(c) Heiko Purnhagen 1993-2000
e-mail: purnhage@tnt.uni-hannover.de
WWW:    http://www.fet.uni-hannover.de/purnhage/

940103   previous version
990823   bp7/dpmi modifications
000115   dmpi for win

************************************************************
}

interface

uses
  dos;

type
  cursort = array [0..31] of word;

var
  cursor : cursort;

function mouseinit : boolean;
procedure showcursor;
procedure hidecursor;
procedure getposbut (var x,y : integer; var bl,br : boolean);
procedure setpos (x,y : integer);
procedure setminmaxx (minx,maxx : integer);
procedure setminmaxy (miny,maxy : integer);
procedure defgcursor (xhot,yhot : integer);
procedure setcursordef;
procedure setcursordefc;


implementation

{$ifndef msdos}
uses
  objects,winapi,_dpmi;

type
  tdoublerec = record
    case byte of
    0:(selector,segment:word;);
    1:(l:longint;);
  end;
{$endif}

var
  regsm : registers;
{$ifndef msdos}
  cursorptr : ^cursort;
  tregs : tcallstructure;
  dosbuffer : tdoublerec;
  mouseprocaddr : pointer;
{$endif}


function mouseinit : boolean;
begin
  regsm.ax := 0;
  intr ($33,regsm);
  mouseinit := (regsm.ax = $ffff) and ((regsm.bx = 2) or (regsm.bx = 3));
end;

procedure showcursor;
begin
  regsm.ax := 1;
  intr ($33,regsm);
end;

procedure hidecursor;
begin
  regsm.ax := 2;
  intr ($33,regsm);
end;

procedure getposbut (var x,y : integer; var bl,br : boolean);
begin
  regsm.ax := 3;
  intr ($33,regsm);
  x := regsm.cx;
  y := regsm.dx;
  bl := (regsm.bx and 1 = 1);
  br := (regsm.bx and 2 = 2);
end;

procedure setpos (x,y : integer);
begin
  regsm.ax := 4;
  regsm.cx := x;
  regsm.dx := y;
  intr ($33,regsm);
end;

procedure setminmaxx (minx,maxx : integer);
begin
  regsm.ax := 7;
  regsm.cx := minx;
  regsm.dx := maxx;
  intr ($33,regsm);
end;

procedure setminmaxy (miny,maxy : integer);
begin
  regsm.ax := 8;
  regsm.cx := miny;
  regsm.dx := maxy;
  intr ($33,regsm);
end;

{$ifdef msdos}
procedure defgcursor (xhot,yhot : integer);
begin
  regsm.ax := 9;
  regsm.bx := xhot;
  regsm.cx := yhot;
  regsm.dx := ofs(cursor);
  regsm.es := seg(cursor);
  intr ($33,regsm);
end;
{$else}
procedure defgcursor (xhot,yhot : integer);
begin
  cursorptr^ := cursor;
  tregs._eax := 9;
  tregs._ebx := xhot;
  tregs._ecx := yhot;
  tregs._edx := 0;
  tregs._es := dosbuffer.segment;
  tregs._ds := 0;
  tregs._cs := ptrrec(mouseprocaddr).seg;
  tregs._ip := ptrrec(mouseprocaddr).ofs;
  callrealmproc(tregs);
end;
{$endif}


procedure setcursordef;
begin
  cursor[0] := $3fff;
  cursor[1] := $1fff;
  cursor[2] := $0fff;
  cursor[3] := $07ff;
  cursor[4] := $03ff;
  cursor[5] := $01ff;
  cursor[6] := $00ff;
  cursor[7] := $007f;
  cursor[8] := $003f;
  cursor[9] := $001f;
  cursor[10] := $001f;
  cursor[11] := $00ff;
  cursor[12] := $10ff;
  cursor[13] := $f87f;
  cursor[14] := $f87f;
  cursor[15] := $f87f;
  cursor[16] := $0000;
  cursor[17] := $4000;
  cursor[18] := $6000;
  cursor[19] := $7000;
  cursor[20] := $7800;
  cursor[21] := $7c00;
  cursor[22] := $7e00;
  cursor[23] := $7f00;
  cursor[24] := $7f80;
  cursor[25] := $7fc0;
  cursor[26] := $7c00;
  cursor[27] := $4600;
  cursor[28] := $0600;
  cursor[29] := $0300;
  cursor[30] := $0300;
  cursor[31] := $0000;
end;

procedure setcursordefc;
begin
  cursor[0] := $0fff;
  cursor[1] := $03ff;
  cursor[2] := $00ff;
  cursor[3] := $00ff;
  cursor[4] := $003f;
  cursor[5] := $003f;
  cursor[6] := $000f;
  cursor[7] := $000f;
  cursor[8] := $0003;
  cursor[9] := $0003;
  cursor[10] := $0003;
  cursor[11] := $000f;
  cursor[12] := $f00f;
  cursor[13] := $fc0f;
  cursor[14] := $fc0f;
  cursor[15] := $ffff;
  cursor[16] := $0000;
  cursor[17] := $3000;
  cursor[18] := $3800;
  cursor[19] := $3c00;
  cursor[20] := $3e00;
  cursor[21] := $3f00;
  cursor[22] := $3f80;
  cursor[23] := $3fc0;
  cursor[24] := $3fe0;
  cursor[25] := $3f30;
  cursor[26] := $3300;
  cursor[27] := $0380;
  cursor[28] := $02c0;
  cursor[29] := $00c0;
  cursor[30] := $0000;
  cursor[31] := $0000;
end;

begin
{$ifndef msdos}
  { allocate dos memory in dpmi/win mode }
  dosbuffer.l := GlobalDosAlloc(sizeof(cursort));
  if dosbuffer.l = 0 then   { error }
  begin
    writeln('mousedrv dos memory alloc error!');
    halt;
  end;
  cursorptr := ptr(dosbuffer.selector,0);
  getrealmintvec($33,mouseprocaddr);
{
writeln('seg=',dosbuffer.segment,' sel=',dosbuffer.selector);
readln;
}
{$endif}
end.

{ end unit }
